# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from .base import Num2Word_Base
from .currency import parse_currency_parts


class Num2Word_KO(Num2Word_Base):
    CURRENCY_FORMS = {
        'KRW': ('원', None),
        'USD': ('달러', '센트'),
        'JPY': ('엔', None)
    }

    def set_high_numwords(self, high):
        max = 4 * len(high)
        for word, n in zip(high, range(max, 0, -4)):
            self.cards[10 ** n] = word

    def setup(self):
        super(Num2Word_KO, self).setup()

        self.negword = "마이너스 "
        self.pointword = "점"

        self.high_numwords = [
            '무량대수',
            '불가사의',
            '나유타',
            '아승기',
            '항하사',
            '극',
            '재',
            '정',
            '간',
            '구',
            '양',
            '자',
            '해',
            '경',
            '조',
            '억',
            '만']
        self.mid_numwords = [(1000, "천"), (100, "백")]
        self.low_numwords = ["십", "구", "팔", "칠", "육", "오", "사", "삼", "이",
                             "일", "영"]
        self.ords = {"일": "한",
                     "이": "두",
                     "삼": "세",
                     "사": "네",
                     "오": "다섯",
                     "육": "여섯",
                     "칠": "일곱",
                     "팔": "여덟",
                     "구": "아홉",
                     "십": "열",
                     "이십": "스물",
                     "삼십": "서른",
                     "사십": "마흔",
                     "오십": "쉰",
                     "육십": "예순",
                     "칠십": "일흔",
                     "팔십": "여든",
                     "구십": "아흔"}

    def merge(self, lpair, rpair):
        ltext, lnum = lpair
        rtext, rnum = rpair
        if lnum == 1 and rnum <= 10000:
            return rpair
        elif 10000 > lnum > rnum:
            return ("%s%s" % (ltext, rtext), lnum + rnum)
        elif lnum >= 10000 and lnum > rnum:
            return ("%s %s" % (ltext, rtext), lnum + rnum)
        else:
            return ("%s%s" % (ltext, rtext), lnum * rnum)

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        if value == 1:
            return "첫 번째"
        outwords = self.to_cardinal(value).split(" ")
        lastwords = outwords[-1].split("백")
        if "십" in lastwords[-1]:
            ten_one = lastwords[-1].split("십")
            ten_one[0] = self.ords[ten_one[0] + "십"]
            try:
                ten_one[1] = self.ords[ten_one[1]]
                ten_one[0] = ten_one[0].replace("스무", "스물")
            except KeyError:
                pass
            lastwords[-1] = ''.join(ten_one)
        else:
            lastwords[-1] = self.ords[lastwords[-1]]
        outwords[-1] = "백 ".join(lastwords)
        return " ".join(outwords) + " 번째"

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return "%s 번째" % (value)

    def to_year(self, val, suffix=None, longval=True):
        if val < 0:
            val = abs(val)
            suffix = '기원전' if not suffix else suffix
        valtext = self.to_cardinal(val)
        return ("%s년" % valtext if not suffix
                else "%s %s년" % (suffix, valtext))

    def to_currency(self, val, currency="KRW", cents=False, separator="",
                    adjective=False):
        left, right, is_negative = parse_currency_parts(
            val, is_int_with_cents=cents)

        try:
            cr1, cr2 = self.CURRENCY_FORMS[currency]
            if (cents or right) and not cr2:
                raise ValueError('Decimals not supported for "%s"' % currency)
        except KeyError:
            raise NotImplementedError(
                'Currency code "%s" not implemented for "%s"' %
                (currency, self.__class__.__name__))

        minus_str = self.negword if is_negative else ""
        return '%s%s%s%s%s' % (
            minus_str,
            ''.join(self.to_cardinal(left).split()),
            cr1,
            ' ' + self.to_cardinal(right)
            if cr2 else '',
            cr2 if cr2 else '',
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

import math

from .lang_EU import Num2Word_EU

GENERIC_DOLLARS = ('dólar', 'dólares')
GENERIC_CENTS = ('centavo', 'centavos')
CURRENCIES_UNA = ('SLL', 'SEK', 'NOK', 'CZK', 'DKK', 'ISK',
                  'SKK', 'GBP', 'CYP', 'EGP', 'FKP', 'GIP',
                  'LBP', 'SDG', 'SHP', 'SSP', 'SYP', 'INR',
                  'IDR', 'LKR', 'MUR', 'NPR', 'PKR', 'SCR',
                  'ESP', 'TRY', 'ITL')
CENTS_UNA = ('EGP', 'JOD', 'LBP', 'SDG', 'SSP', 'SYP')


class Num2Word_ES(Num2Word_EU):
    CURRENCY_FORMS = {
        'EUR': (('euro', 'euros'), ('céntimo', 'céntimos')),
        'ESP': (('peseta', 'pesetas'), ('céntimo', 'céntimos')),
        'USD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'PEN': (('sol', 'soles'), ('céntimo', 'céntimos')),
        'CRC': (('colón', 'colones'), GENERIC_CENTS),
        'AUD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'CAD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'GBP': (('libra', 'libras'), ('penique', 'peniques')),
        'RUB': (('rublo', 'rublos'), ('kopeyka', 'kopeykas')),
        'SEK': (('corona', 'coronas'), ('öre', 'öre')),
        'NOK': (('corona', 'coronas'), ('øre', 'øre')),
        'PLN': (('zloty', 'zlotys'), ('grosz', 'groszy')),
        'MXN': (('peso', 'pesos'), GENERIC_CENTS),
        'RON': (('leu', 'leus'), ('ban', 'bani')),
        'INR': (('rupia', 'rupias'), ('paisa', 'paisas')),
        'HUF': (('florín', 'florines'), ('fillér', 'fillér')),
        'FRF': (('franco', 'francos'), ('céntimo', 'céntimos')),
        'CNY': (('yuan', 'yuanes'), ('fen', 'jiaos')),
        'CZK': (('corona', 'coronas'), ('haléř', 'haléř')),
        'NIO': (('córdoba', 'córdobas'), GENERIC_CENTS),
        'VES': (('bolívar', 'bolívares'), ('céntimo', 'céntimos')),
        'BRL': (('real', 'reales'), GENERIC_CENTS),
        'CHF': (('franco', 'francos'), ('céntimo', 'céntimos')),
        'JPY': (('yen', 'yenes'), ('sen', 'sen')),
        'KRW': (('won', 'wones'), ('jeon', 'jeon')),
        'KPW': (('won', 'wones'), ('chon', 'chon')),
        'TRY': (('lira', 'liras'), ('kuruş', 'kuruş')),
        'ZAR': (('rand', 'rands'), ('céntimo', 'céntimos')),
        'KZT': (('tenge', 'tenges'), ('tïın', 'tïın')),
        'UAH': (('hryvnia', 'hryvnias'), ('kopiyka', 'kopiykas')),
        'THB': (('baht', 'bahts'), ('satang', 'satang')),
        'AED': (('dirham', 'dirhams'), ('fils', 'fils')),
        'AFN': (('afghani', 'afghanis'), ('pul', 'puls')),
        'ALL': (('lek ', 'leke'), ('qindarkë', 'qindarka')),
        'AMD': (('dram', 'drams'), ('luma', 'lumas')),
        'ANG': (('florín', 'florines'), GENERIC_CENTS),
        'AOA': (('kwanza', 'kwanzas'), ('céntimo', 'céntimos')),
        'ARS': (('peso', 'pesos'), GENERIC_CENTS),
        'AWG': (('florín', 'florines'), GENERIC_CENTS),
        'AZN': (('manat', 'manat'), ('qəpik', 'qəpik')),
        'BBD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'BDT': (('taka', 'takas'), ('paisa', 'paisas')),
        'BGN': (('lev', 'leva'), ('stotinka', 'stotinki')),
        'BHD': (('dinar', 'dinares'), ('fils', 'fils')),
        'BIF': (('franco', 'francos'), ('céntimo', 'céntimos')),
        'BMD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'BND': (GENERIC_DOLLARS, GENERIC_CENTS),
        'BOB': (('boliviano', 'bolivianos'), GENERIC_CENTS),
        'BSD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'BTN': (('ngultrum', 'ngultrum'), ('chetrum', 'chetrum')),
        'BWP': (('pula', 'pulas'), ('thebe', 'thebes')),
        'BYN': (('rublo', 'rublos'), ('kópek', 'kópeks')),
        'BYR': (('rublo', 'rublos'), ('kópek', 'kópeks')),
        'BZD': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'CDF': (('franco', 'francos'), ('céntimo', 'céntimos')),
        'CLP': (('peso', 'pesos'), GENERIC_CENTS),
        'COP': (('peso', 'pesos'), GENERIC_CENTS),
        'CUP': (('peso', 'pesos'), GENERIC_CENTS),
        'CVE': (('escudo', 'escudos'), GENERIC_CENTS),
        'CYP': (('libra', 'libras'), ('céntimo', 'céntimos')),
        'DJF': (('franco', 'francos'), ('céntimo', 'céntimos')),
        'DKK': (('corona', 'coronas'), ('øre', 'øre')),
        'DOP': (('peso', 'pesos'), GENERIC_CENTS),
        'DZD': (('dinar', 'dinares'), ('céntimo', 'céntimos')),
        'ECS': (('sucre', 'sucres'), GENERIC_CENTS),
        'EGP': (('libra', 'libras'), ('piastra', 'piastras')),
        'ERN': (('nakfa', 'nakfas'), ('céntimo', 'céntimos')),
        'ETB': (('birr', 'birrs'), ('céntimo', 'céntimos')),
        'FJD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'FKP': (('libra', 'libras'), ('penique', 'peniques')),
        'GEL': (('lari', 'laris'), ('tetri', 'tetris')),
        'GHS': (('cedi', 'cedis'), ('pesewa', 'pesewas')),
        'GIP': (('libra', 'libras'), ('penique', 'peniques')),
        'GMD': (('dalasi', 'dalasis'), ('butut', 'bututs')),
        'GNF': (('franco', 'francos'), ('céntimo', 'céntimos')),
        'GTQ': (('quetzal', 'quetzales'), GENERIC_CENTS),
        'GYD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'HKD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'HNL': (('lempira', 'lempiras'), GENERIC_CENTS),
        'HRK': (('kuna', 'kunas'), ('lipa', 'lipas')),
        'HTG': (('gourde', 'gourdes'), ('céntimo', 'céntimos')),
        'IDR': (('rupia', 'rupias'), ('céntimo', 'céntimos')),
        'ILS': (('séquel', 'séqueles'), ('agora', 'agoras')),
        'IQD': (('dinar', 'dinares'), ('fils', 'fils')),
        'IRR': (('rial', 'riales'), ('dinar', 'dinares')),
        'ISK': (('corona', 'coronas'), ('eyrir', 'aurar')),
        'ITL': (('lira', 'liras'), ('céntimo', 'céntimos')),
        'JMD': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'JOD': (('dinar', 'dinares'), ('piastra', 'piastras')),
        'KES': (('chelín', 'chelines'), ('céntimo', 'céntimos')),
        'KGS': (('som', 'som'), ('tyiyn', 'tyiyn')),
        'KHR': (('riel', 'rieles'), ('céntimo', 'céntimos')),
        'KMF': (('franco', 'francos'), ('céntimo', 'céntimos')),
        'KWD': (('dinar', 'dinares'), ('fils', 'fils')),
        'KYD': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'LAK': (('kip', 'kips'), ('att', 'att')),
        'LBP': (('libra', 'libras'), ('piastra', 'piastras')),
        'LKR': (('rupia', 'rupias'), ('céntimo', 'céntimos')),
        'LRD': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'LSL': (('loti', 'lotis'), ('céntimo', 'céntimos')),
        'LTL': (('lita', 'litas'), ('céntimo', 'céntimos')),
        'LVL': (('lat', 'lats'), ('céntimo', 'céntimos')),
        'LYD': (('dinar', 'dinares'), ('dírham', 'dírhams')),
        'MAD': (('dírham', 'dirhams'), ('céntimo', 'céntimos')),
        'MDL': (('leu', 'lei'), ('ban', 'bani')),
        'MGA': (('ariary', 'ariaris'), ('iraimbilanja', 'iraimbilanja')),
        'MKD': (('denar', 'denares'), ('deni', 'denis')),
        'MMK': (('kiat', 'kiats'), ('pya', 'pyas')),
        'MNT': (('tugrik', 'tugriks'), ('möngö', 'möngö')),
        'MOP': (('pataca', 'patacas'), ('avo', 'avos')),
        'MRO': (('ouguiya', 'ouguiyas'), ('khoums', 'khoums')),
        'MRU': (('ouguiya', 'ouguiyas'), ('khoums', 'khoums')),
        'MUR': (('rupia', 'rupias'), ('céntimo', 'céntimos')),
        'MVR': (('rufiyaa', 'rufiyaas'), ('laari', 'laari')),
        'MWK': (('kuacha', 'kuachas'), ('tambala', 'tambalas')),
        'MYR': (('ringgit', 'ringgit'), ('céntimo', 'céntimos')),
        'MZN': (('metical', 'metical'), GENERIC_CENTS),
        'NAD': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'NGN': (('naira', 'nairas'), ('kobo', 'kobo')),
        'NPR': (('rupia', 'rupias'), ('paisa', 'paisas')),
        'NZD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'OMR': (('rial', 'riales'), ('baisa', 'baisa')),
        'PAB': (('balboa', 'balboas'), ('centésimo', 'centésimos')),
        'PGK': (('kina', 'kinas'), ('toea', 'toea')),
        'PHP': (('peso', 'pesos'), GENERIC_CENTS),
        'PKR': (('rupia', 'rupias'), ('paisa', 'paisas')),
        'PLZ': (('zloty', 'zlotys'), ('grosz', 'groszy')),
        'PYG': (('guaraní', 'guaranís'), ('céntimo', 'céntimos')),
        'QAR': (('rial', 'riales'), ('dírham', 'dírhams')),
        'QTQ': (('quetzal', 'quetzales'), GENERIC_CENTS),
        'RSD': (('dinar', 'dinares'), ('para', 'para')),
        'RUR': (('rublo', 'rublos'), ('kopek', 'kopeks')),
        'RWF': (('franco', 'francos'), ('céntimo', 'céntimos')),
        'SAR': (('riyal', 'riales'), ('halala', 'halalas')),
        'SBD': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'SCR': (('rupia', 'rupias'), ('céntimo', 'céntimos')),
        'SDG': (('libra', 'libras'), ('piastra', 'piastras')),
        'SGD': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'SHP': (('libra', 'libras'), ('penique', 'peniques')),
        'SKK': (('corona', 'coronas'), ('halier', 'haliers')),
        'SLL': (('leona', 'leonas'), ('céntimo', 'céntimos')),
        'SRD': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'SSP': (('libra', 'libras'), ('piastra', 'piastras')),
        'STD': (('dobra', 'dobras'), ('céntimo', 'céntimos')),
        'SVC': (('colón', 'colones'), GENERIC_CENTS),
        'SYP': (('libra', 'libras'), ('piastra', 'piastras')),
        'SZL': (('lilangeni', 'emalangeni'), ('céntimo', 'céntimos')),
        'TJS': (('somoni', 'somonis'), ('dirame', 'dirames')),
        'TMT': (('manat', 'manat'), ('tenge', 'tenge')),
        'TND': (('dinar', 'dinares'), ('milésimo', 'milésimos')),
        'TOP': (('paanga', 'paangas'), ('céntimo', 'céntimos')),
        'TTD': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'TWD': (('nuevo dólar', 'nuevos dólares'), ('céntimo', 'céntimos')),
        'TZS': (('chelín', 'chelines'), ('céntimo', 'céntimos')),
        'UAG': (('hryvnia', 'hryvnias'), ('kopiyka', 'kopiykas')),
        'UGX': (('chelín', 'chelines'), ('céntimo', 'céntimos')),
        'UYU': (('peso', 'pesos'), ('centésimo', 'centésimos')),
        'UZS': (('sum', 'sum'), ('tiyin', 'tiyin')),
        'VEF': (('bolívar fuerte', 'bolívares fuertes'),
                ('céntimo', 'céntimos')),
        'VND': (('dong', 'dongs'), ('xu', 'xu')),
        'VUV': (('vatu', 'vatu'), ('nenhum', 'nenhum')),
        'WST': (('tala', 'tala'), GENERIC_CENTS),
        'XAF': (('franco CFA', 'francos CFA'), ('céntimo', 'céntimos')),
        'XCD': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'XOF': (('franco CFA', 'francos CFA'), ('céntimo', 'céntimos')),
        'XPF': (('franco CFP', 'francos CFP'), ('céntimo', 'céntimos')),
        'YER': (('rial', 'riales'), ('fils', 'fils')),
        'YUM': (('dinar', 'dinares'), ('para', 'para')),
        'ZMW': (('kwacha', 'kwachas'), ('ngwee', 'ngwee')),
        'ZRZ': (('zaire', 'zaires'), ('likuta', 'makuta')),
        'ZWL': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
        'ZWL': (GENERIC_DOLLARS, ('céntimo', 'céntimos')),
    }

    # //CHECK: Is this sufficient??
    GIGA_SUFFIX = None
    MEGA_SUFFIX = "illón"

    def setup(self):
        lows = ["cuatr", "tr", "b", "m"]
        self.high_numwords = self.gen_high_numwords([], [], lows)
        self.negword = "menos "
        self.pointword = "punto"
        self.errmsg_nonnum = "type(%s) no es [long, int, float]"
        self.errmsg_floatord = "El float %s no puede ser tratado como un" \
            " ordinal."
        self.errmsg_negord = "El número negativo %s no puede ser tratado" \
            " como un ordinal."
        self.errmsg_toobig = (
            "abs(%s) deber ser inferior a %s."
            )
        self.gender_stem = "o"
        self.exclude_title = ["y", "menos", "punto"]
        self.mid_numwords = [(1000, "mil"), (100, "cien"), (90, "noventa"),
                             (80, "ochenta"), (70, "setenta"), (60, "sesenta"),
                             (50, "cincuenta"), (40, "cuarenta"),
                             (30, "treinta")]
        self.low_numwords = ["veintinueve", "veintiocho", "veintisiete",
                             "veintiséis", "veinticinco", "veinticuatro",
                             "veintitrés", "veintidós", "veintiuno",
                             "veinte", "diecinueve", "dieciocho", "diecisiete",
                             "dieciséis", "quince", "catorce", "trece", "doce",
                             "once", "diez", "nueve", "ocho", "siete", "seis",
                             "cinco", "cuatro", "tres", "dos", "uno", "cero"]
        self.ords = {1: "primer",
                     2: "segund",
                     3: "tercer",
                     4: "cuart",
                     5: "quint",
                     6: "sext",
                     7: "séptim",
                     8: "octav",
                     9: "noven",
                     10: "décim",
                     20: "vigésim",
                     30: "trigésim",
                     40: "quadragésim",
                     50: "quincuagésim",
                     60: "sexagésim",
                     70: "septuagésim",
                     80: "octogésim",
                     90: "nonagésim",
                     100: "centésim",
                     200: "ducentésim",
                     300: "tricentésim",
                     400: "cuadrigentésim",
                     500: "quingentésim",
                     600: "sexcentésim",
                     700: "septigentésim",
                     800: "octigentésim",
                     900: "noningentésim",
                     1e3: "milésim",
                     1e6: "millonésim",
                     1e9: "billonésim",
                     1e12: "trillonésim",
                     1e15: "cuadrillonésim"}

    def merge(self, curr, next):
        ctext, cnum, ntext, nnum = curr + next

        if cnum == 1:
            if nnum < 1000000:
                return next
            ctext = "un"
        elif cnum == 100 and not nnum % 1000 == 0:
            ctext += "t" + self.gender_stem

        if nnum < cnum:
            if cnum < 100:
                return "%s y %s" % (ctext, ntext), cnum + nnum
            return "%s %s" % (ctext, ntext), cnum + nnum
        elif (not nnum % 1000000) and cnum > 1:
            ntext = ntext[:-3] + "lones"

        if nnum == 100:
            if cnum == 5:
                ctext = "quinien"
                ntext = ""
            elif cnum == 7:
                ctext = "sete"
            elif cnum == 9:
                ctext = "nove"
            ntext += "t" + self.gender_stem + "s"
        else:
            ntext = " " + ntext

        return (ctext + ntext, cnum * nnum)

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        if value == 0:
            text = ""
        elif value <= 10:
            text = "%s%s" % (self.ords[value], self.gender_stem)
        elif value <= 12:
            text = (
                "%s%s%s" % (self.ords[10], self.gender_stem,
                            self.to_ordinal(value - 10))
                    )
        elif value <= 100:
            dec = (value // 10) * 10
            text = (
                "%s%s %s" % (self.ords[dec], self.gender_stem,
                             self.to_ordinal(value - dec))
                    )
        elif value <= 1e3:
            cen = (value // 100) * 100
            text = (
                "%s%s %s" % (self.ords[cen], self.gender_stem,
                             self.to_ordinal(value - cen))
                    )
        elif value < 1e18:
            # Round down to the nearest 1e(3n)
            # dec contains the following:
            # [ 1e3,  1e6): 1e3
            # [ 1e6,  1e9): 1e6
            # [ 1e9, 1e12): 1e9
            # [1e12, 1e15): 1e12
            # [1e15, 1e18): 1e15
            dec = 1000 ** int(math.log(int(value), 1000))

            # Split the parts before and after the word for 'dec'
            # eg (12, 345) = divmod(12_345, 1_000)
            high_part, low_part = divmod(value, dec)

            cardinal = self.to_cardinal(high_part) if high_part != 1 else ""
            text = (
                "%s%s%s %s" % (cardinal, self.ords[dec], self.gender_stem,
                               self.to_ordinal(low_part))
                    )
        else:
            text = self.to_cardinal(value)
        return text.strip()

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return "%s%s" % (value, "º" if self.gender_stem == 'o' else "ª")

    def to_currency(self, val, currency='EUR', cents=True, separator=' con',
                    adjective=False):
        result = super(Num2Word_ES, self).to_currency(
            val, currency=currency, cents=cents, separator=separator,
            adjective=adjective)
        # Handle exception: In Spanish it's "un euro" and not "uno euro",
        # except in these currencies, where it's "una": leona, corona,
        # libra, lira, rupia, lempira, peseta.
        # The same goes for "veintiuna", "treinta y una"...
        # Also, this needs to be handled separately for "dollars" and
        # "cents".
        # All "cents" are masculine except for: piastra.
        # Source: https://www.rae.es/dpd/una (section 2.2)

        # split "dollars" part from "cents" part
        list_result = result.split(separator + " ")

        # "DOLLARS" PART (list_result[0])

        # Feminine currencies ("una libra", "trescientas libras"...)
        if currency in CURRENCIES_UNA:

            # "una libra", "veintiuna libras", "treinta y una libras"...
            list_result[0] = list_result[0].replace("uno", "una")

            # "doscientas libras", "trescientas libras"...
            list_result[0] = list_result[0].replace("cientos", "cientas")

        # Masc.: Correct orthography for the specific case of "veintiún":
        list_result[0] = list_result[0].replace("veintiuno", "veintiún")

        # Masculine currencies: general case ("un euro", "treinta y un
        # euros"...):
        list_result[0] = list_result[0].replace("uno", "un")

        # "CENTS" PART (list_result[1])

        # Feminine "cents" ("una piastra", "veintiuna piastras"...)
        if currency in CENTS_UNA:

            # "una piastra", "veintiuna piastras", "treinta y una piastras"...
            list_result[1] = list_result[1].replace("uno", "una")

        # Masc.: Correct orthography for the specific case of "veintiún":
        list_result[1] = list_result[1].replace("veintiuno", "veintiún")

        # Masculine "cents": general case ("un centavo", "treinta y un
        # centavos"...):
        list_result[1] = list_result[1].replace("uno", "un")

        # join back "dollars" part with "cents" part
        result = (separator + " ").join(list_result)

        return result



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from . import lang_EU

# Genders
KK = 0  # Karlkyn (male)
KVK = 1  # Kvenkyn (female)
HK = 2  # Hvorugkyn (neuter)

GENDERS = {
    "einn": ("einn", "ein", "eitt"),
    "tveir": ("tveir", "tvær", "tvö"),
    "þrír": ("þrír", "þrjár", "þrjú"),
    "fjórir": ("fjórir", "fjórar", "fjögur"),
}

PLURALS = {
    "hundrað": ("hundrað", "hundruð"),
}


class Num2Word_IS(lang_EU.Num2Word_EU):

    GIGA_SUFFIX = "illjarður"
    MEGA_SUFFIX = "illjón"

    def setup(self):
        lows = ["okt", "sept", "sext", "kvint", "kvaðr", "tr", "b", "m"]
        self.high_numwords = self.gen_high_numwords([], [], lows)

        self.negword = "mínus "
        self.pointword = "komma"

        # All words should be excluded, title case is not used in Icelandic
        self.exclude_title = ["og", "komma", "mínus"]

        self.mid_numwords = [(1000, "þúsund"), (100, "hundrað"),
                             (90, "níutíu"), (80, "áttatíu"), (70, "sjötíu"),
                             (60, "sextíu"), (50, "fimmtíu"), (40, "fjörutíu"),
                             (30, "þrjátíu")]
        self.low_numwords = ["tuttugu", "nítján", "átján", "sautján",
                             "sextán", "fimmtán", "fjórtán", "þrettán",
                             "tólf", "ellefu", "tíu", "níu", "átta",
                             "sjö", "sex", "fimm", "fjórir", "þrír",
                             "tveir", "einn", "núll"]
        self.ords = {"einn": "fyrsti",
                     "tveir": "annar",
                     "þrír": "þriðji",
                     "fjórir": "fjórði",
                     "fimm": "fimmti",
                     "sex": "sjötti",
                     "sjö": "sjöundi",
                     "átta": "áttundi",
                     "níu": "níundi",
                     "tíu": "tíundi",
                     "ellefu": "ellefti",
                     "tólf": "tólfti"}

    def pluralize(self, n, noun):
        form = 0 if (n % 10 == 1 and n % 100 != 11) else 1
        if form == 0:
            return noun
        elif self.GIGA_SUFFIX in noun:
            return noun.replace(self.GIGA_SUFFIX, "illjarðar")
        elif self.MEGA_SUFFIX in noun:
            return noun.replace(self.MEGA_SUFFIX, "illjónir")
        elif noun not in PLURALS:
            return noun
        return PLURALS[noun][form]

    def genderize(self, adj, noun):
        last = adj.split()[-1]
        if last not in GENDERS:
            return adj
        gender = KK
        if "hund" in noun or "þús" in noun:
            gender = HK
        elif "illjarð" in noun:
            gender = KK
        elif "illjón" in noun:
            gender = KVK
        return adj.replace(last, GENDERS[last][gender])

    def merge(self, lpair, rpair):
        ltext, lnum = lpair
        rtext, rnum = rpair

        if lnum == 1 and rnum < 100:
            return (rtext, rnum)
        elif lnum < rnum:
            rtext = self.pluralize(lnum, rtext)
            ltext = self.genderize(ltext, rtext)
            return ("%s %s" % (ltext, rtext), lnum * rnum)
        elif lnum > rnum and rnum in self.cards:
            rtext = self.pluralize(lnum, rtext)
            ltext = self.genderize(ltext, rtext)
            return ("%s og %s" % (ltext, rtext), lnum + rnum)
        return ("%s %s" % (ltext, rtext), lnum + rnum)

    def to_ordinal(self, value):
        raise NotImplementedError

    def to_ordinal_num(self, value):
        raise NotImplementedError

    def to_year(self, val, suffix=None, longval=True):
        raise NotImplementedError

    def to_currency(self, val, longval=True):
        raise NotImplementedError



# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.
# Copyright (c) 2022, Sergei Ruzki.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .base import Num2Word_Base
from .utils import get_digits, splitbyx

ZERO = "нуль"

ONES_FEMININE = {
    1: "адна",
    2: "дзве",
    3: "тры",
    4: "чатыры",
    5: "пяць",
    6: "шэсць",
    7: "сем",
    8: "восем",
    9: "дзевяць",
}

ONES = {
    "f": {
        1: "адна",
        2: "дзве",
        3: "тры",
        4: "чатыры",
        5: "пяць",
        6: "шэсць",
        7: "сем",
        8: "восем",
        9: "дзевяць",
    },
    "m": {
        1: "адзін",
        2: "два",
        3: "тры",
        4: "чатыры",
        5: "пяць",
        6: "шэсць",
        7: "сем",
        8: "восем",
        9: "дзевяць",
    },
    "n": {
        1: "адно",
        2: "два",
        3: "тры",
        4: "чатыры",
        5: "пяць",
        6: "шэсць",
        7: "сем",
        8: "восем",
        9: "дзевяць",
    },
}

TENS = {
    0: "дзесяць",
    1: "адзінаццаць",
    2: "дванаццаць",
    3: "трынаццаць",
    4: "чатырнаццаць",
    5: "пятнаццаць",
    6: "шаснаццаць",
    7: "сямнаццаць",
    8: "васямнаццаць",
    9: "дзевятнаццаць",
}

TWENTIES = {
    2: "дваццаць",
    3: "трыццаць",
    4: "сорак",
    5: "пяцьдзясят",
    6: "шэсцьдзясят",
    7: "семдзесят",
    8: "восемдзесят",
    9: "дзевяноста",
}

TWENTIES_ORD = (
    ("дваццаць", "дваццаці"),
    ("трыццаць", "трыццаці"),
    ("сорак", "сарака"),
    ("пяцьдзясят", "пяцідзясяці"),
    ("шэсцьдзясят", "шaсцідзясяці"),
    ("семдзесят", "сямідзесяці"),
    ("восемдзесят", "васьмідзесяці"),
    ("дзевяноста", "дзевяноста"),
)

HUNDREDS = {
    1: "сто",
    2: "дзвесце",
    3: "трыста",
    4: "чатырыста",
    5: "пяцьсот",
    6: "шэсцьсот",
    7: "семсот",
    8: "восемсот",
    9: "дзевяцьсот",
}

THOUSANDS = {
    1: ("тысяча", "тысячы", "тысяч"),  # 10^3
    2: ("мільён", "мільёны", "мільёнаў"),  # 10^6
    3: ("мільярд", "мільярды", "мільярдаў"),  # 10^9
    4: ("трыльён", "трыльёны", "трыльёнаў"),  # 10^12
    5: ("квадрыльён", "квадрыльёны", "квадрыльёнаў"),  # 10^15
    6: ("квінтыльён", "квінтыльёны", "квінтыльёнаў"),  # 10^18
    7: ("секстыльён", "секстыльёны", "секстыльёнаў"),  # 10^21
    8: ("сэптыльён", "сэптыльёны", "сэптыльёнаў"),  # 10^24
    9: ("актыльён", "актыльёны", "актыльёнаў"),  # 10^27
    10: ("нанільён", "нанільёны", "нанільёнаў"),  # 10^30
}


class Num2Word_BE(Num2Word_Base):
    CURRENCY_FORMS = {
        "RUB": (
            ("расійскі рубель", "расійскія рублі", "расійскіх рублёў"),
            ("капейка", "капейкі", "капеек"),
        ),
        "EUR": (("еўра", "еўра", "еўра"), ("цэнт", "цэнты", "цэнтаў")),
        "USD": (("долар", "долары", "долараў"), ("цэнт", "цэнты", "цэнтаў")),
        "UAH": (
            ("грыўна", "грыўны", "грыўнаў"),
            ("капейка", "капейкі", "капеек"),
        ),
        "KZT": (("тэнге", "тэнге", "тэнге"), ("тыйін", "тыйіны", "тыйінаў")),
        "BYN": (
            ("беларускі рубель", "беларускія рублі", "беларускіх рублёў"),
            ("капейка", "капейкі", "капеек"),
        ),
        "UZS": (("сум", "сумы", "сумаў"), ("тыйін", "тыйіны", "тыйінаў")),
        "PLN": (("злоты", "злотых", "злотых"), ("грош", "грошы", "грошаў")),
    }

    def setup(self):
        self.negword = "мінус"
        self.pointword = "коска"
        self.ords = {
            "нуль": "нулявы",
            "адзін": "першы",
            "два": "другі",
            "тры": "трэці",
            "чатыры": "чацвёрты",
            "пяць": "пяты",
            "шэсць": "шосты",
            "сем": "сёмы",
            "восем": "восьмы",
            "дзевяць": "дзявяты",
            "сто": "соты",
            "тысяча": "тысячны",
        }

        self.ords_adjective = {
            "адзін": "адна",
            "адна": "адна",
            "дзве": "двух",
            "тры": "трох",
            "чатыры": "чатырох",
            "пяць": "пяці",
            "шэсць": "шасці",
            "сем": "сямі",
            "восем": "васьмі",
            "дзевяць": "дзевяцi",
            "сто": "ста",
        }

    def to_cardinal(self, number, gender="m"):
        n = str(number).replace(",", ".")
        if "." in n:
            left, right = n.split(".")
            if set(right) == {"0"}:
                leading_zero_count = 0
            else:
                leading_zero_count = len(right) - len(right.lstrip("0"))

            decimal_part = (ZERO + " ") * leading_zero_count + self._int2word(
                int(right), gender
            )
            return "{} {} {}".format(
                self._int2word(int(left), gender), self.pointword, decimal_part
            )
        else:
            return self._int2word(int(n), gender)

    def pluralize(self, n, forms):
        if n % 100 < 10 or n % 100 > 20:
            if n % 10 == 1:
                form = 0
            elif 5 > n % 10 > 1:
                form = 1
            else:
                form = 2
        else:
            form = 2
        return forms[form]

    def to_ordinal(self, number, gender="m"):
        self.verify_ordinal(number)
        if isinstance(gender, bool) and gender:
            gender = "f"
        outwords = self.to_cardinal(number, gender).split(" ")
        lastword = outwords[-1].lower()
        try:
            if len(outwords) > 1:
                if outwords[-2] in self.ords_adjective:
                    outwords[-2] = self.ords_adjective.get(
                        outwords[-2], outwords[-2]
                    )
                elif outwords[-2] == "дзесяць":
                    outwords[-2] = outwords[-2][:-1] + "і"
            if len(outwords) == 3:
                if outwords[-3] in ["адзін", "адна"]:
                    outwords[-3] = ""
            lastword = self.ords[lastword]
        except KeyError:
            if lastword[:-3] in self.ords_adjective:
                lastword = (
                    self.ords_adjective.get(lastword[:-3], lastword) + "соты"
                )
            elif lastword[-7:] == "дзесяць":
                lastword = "дзясяты"
            elif lastword[-9:] == "семдзесят":
                lastword = "сямідзясяты"
            elif lastword[-1] == "ь" or lastword[-2] == "ц":
                lastword = lastword[:-2] + "ты"
            elif lastword[-1] == "к":
                lastword = lastword.replace("о", "а") + "авы"

            elif lastword[-2] == "ч" or lastword[-1] == "ч":
                if lastword[-2] == "ч":
                    lastword = lastword[:-1] + "ны"
                if lastword[-1] == "ч":
                    lastword = lastword + "ны"

            elif lastword[-1] == "н" or lastword[-2] == "н":
                lastword = lastword[: lastword.rfind("н") + 1] + "ны"
            elif lastword[-3:] == "наў":
                lastword = lastword[: lastword.rfind("н") + 1] + "ны"
            elif lastword[-1] == "д" or lastword[-2] == "д":
                lastword = lastword[: lastword.rfind("д") + 1] + "ны"

        if gender == "f":
            if lastword[-2:] in [
                "ці",
            ]:
                lastword = lastword[:-2] + "цяя"
            else:
                lastword = lastword[:-1] + "ая"

        if gender == "n":
            if lastword[-2:] in ["ці", "ца"]:
                lastword = lastword[:-2] + "цяе"
            else:
                lastword = lastword[:-1] + "ае"

        outwords[-1] = self.title(lastword)
        if len(outwords) == 2 and "адна" in outwords[-2]:
            outwords[-2] = outwords[-1]
            del outwords[-1]

        if len(outwords) > 1 and (
            (any(x[0] in outwords[-1] for x in THOUSANDS.values()))
            or "тысяч" in outwords[-1]
        ):
            new_outwords = []
            for _w in outwords:
                replacement = next(
                    (x for x in TWENTIES_ORD if x[0] in _w), None
                )
                if replacement:
                    _w = _w.replace(replacement[0], replacement[1])
                new_outwords.append(_w)
            outwords = ["".join(new_outwords)]

        return " ".join(outwords).strip()

    def _money_verbose(self, number, currency):
        gender = "m"
        if currency == "UAH":
            gender = "f"

        return self._int2word(number, gender)

    def _cents_verbose(self, number, currency):
        if currency in ("UAH", "RUB", "BYN"):
            gender = "f"
        else:
            gender = "m"

        return self._int2word(number, gender)

    def _int2word(self, n, gender="m"):
        if n < 0:
            return " ".join([self.negword, self._int2word(abs(n), gender)])

        if n == 0:
            return ZERO

        words = []
        chunks = list(splitbyx(str(n), 3))
        i = len(chunks)
        for x in chunks:
            i -= 1

            if x == 0:
                continue

            n1, n2, n3 = get_digits(x)

            if n3 > 0:
                words.append(HUNDREDS[n3])

            if n2 > 1:
                words.append(TWENTIES[n2])

            if n2 == 1:
                words.append(TENS[n1])
            elif n1 > 0:
                if i == 0:
                    ones = ONES[gender]
                elif i == 1:
                    ones = ONES["f"]  # Thousands are feminine
                else:
                    ones = ONES["m"]

                words.append(ones[n1])

            if i > 0:
                words.append(self.pluralize(x, THOUSANDS[i]))

        return " ".join(words)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.
# Copyright (c) 2018, Abdullah Alhazmy, Alhazmy13.  All Rights Reserved.


# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

import decimal
import math
import re
from decimal import Decimal
from math import floor

from .base import Num2Word_Base

CURRENCY_SR = [("ريال", "ريالان", "ريالات", "ريالاً"),
               ("هللة", "هللتان", "هللات", "هللة")]
CURRENCY_EGP = [("جنيه", "جنيهان", "جنيهات", "جنيهاً"),
                ("قرش", "قرشان", "قروش", "قرش")]
CURRENCY_KWD = [("دينار", "ديناران", "دينارات", "ديناراً"),
                ("فلس", "فلسان", "فلس", "فلس")]
CURRENCY_TND = [("دينار", "ديناران", "دينارات", "ديناراً"),
                ("مليماً", "ميلمان", "مليمات", "مليم")]

ARABIC_ONES = [
    "", "واحد", "اثنان", "ثلاثة", "أربعة", "خمسة", "ستة", "سبعة", "ثمانية",
    "تسعة",
    "عشرة", "أحد عشر", "اثنا عشر", "ثلاثة عشر", "أربعة عشر", "خمسة عشر",
    "ستة عشر", "سبعة عشر", "ثمانية عشر",
    "تسعة عشر"
]


class Num2Word_AR(Num2Word_Base):
    errmsg_toobig = "abs(%s) must be less than %s."
    MAXVAL = 10**51

    def __init__(self):
        super().__init__()

        self.number = 0
        self.arabicPrefixText = ""
        self.arabicSuffixText = ""
        self.integer_value = 0
        self._decimalValue = ""
        self.partPrecision = 2
        self.currency_unit = CURRENCY_SR[0]
        self.currency_subunit = CURRENCY_SR[1]
        self.isCurrencyPartNameFeminine = True
        self.isCurrencyNameFeminine = False
        self.separator = 'و'

        self.arabicOnes = ARABIC_ONES
        self.arabicFeminineOnes = [
            "", "إحدى", "اثنتان", "ثلاث", "أربع", "خمس", "ست", "سبع", "ثمان",
            "تسع",
            "عشر", "إحدى عشرة", "اثنتا عشرة", "ثلاث عشرة", "أربع عشرة",
            "خمس عشرة", "ست عشرة", "سبع عشرة", "ثماني عشرة",
            "تسع عشرة"
        ]
        self.arabicOrdinal = [
            "", "اول", "ثاني", "ثالث", "رابع", "خامس", "سادس", "سابع", "ثامن",
            "تاسع", "عاشر", "حادي عشر", "ثاني عشر", "ثالث عشر", "رابع عشر",
            "خامس عشر", "سادس عشر", "سابع عشر", "ثامن عشر", "تاسع عشر"
        ]
        self.arabicTens = [
            "عشرون", "ثلاثون", "أربعون", "خمسون", "ستون", "سبعون", "ثمانون",
            "تسعون"
        ]
        self.arabicHundreds = [
            "", "مائة", "مئتان", "ثلاثمائة", "أربعمائة", "خمسمائة", "ستمائة",
            "سبعمائة", "ثمانمائة", "تسعمائة"
        ]

        self.arabicAppendedTwos = [
            "مئتا", "ألفا", "مليونا", "مليارا", "تريليونا", "كوادريليونا",
            "كوينتليونا", "سكستيليونا", "سبتيليونا", "أوكتيليونا ",
            "نونيليونا", "ديسيليونا", "أندسيليونا", "دوديسيليونا",
            "تريديسيليونا", "كوادريسيليونا", "كوينتينيليونا"
        ]
        self.arabicTwos = [
            "مئتان", "ألفان", "مليونان", "ملياران", "تريليونان",
            "كوادريليونان", "كوينتليونان", "سكستيليونان", "سبتيليونان",
            "أوكتيليونان ", "نونيليونان ", "ديسيليونان", "أندسيليونان",
            "دوديسيليونان", "تريديسيليونان", "كوادريسيليونان", "كوينتينيليونان"
        ]
        self.arabicGroup = [
            "مائة", "ألف", "مليون", "مليار", "تريليون", "كوادريليون",
            "كوينتليون", "سكستيليون", "سبتيليون", "أوكتيليون", "نونيليون",
            "ديسيليون", "أندسيليون", "دوديسيليون", "تريديسيليون",
            "كوادريسيليون", "كوينتينيليون"
        ]
        self.arabicAppendedGroup = [
            "", "ألفاً", "مليوناً", "ملياراً", "تريليوناً", "كوادريليوناً",
            "كوينتليوناً", "سكستيليوناً", "سبتيليوناً", "أوكتيليوناً",
            "نونيليوناً", "ديسيليوناً", "أندسيليوناً", "دوديسيليوناً",
            "تريديسيليوناً", "كوادريسيليوناً", "كوينتينيليوناً"
        ]
        self.arabicPluralGroups = [
            "", "آلاف", "ملايين", "مليارات", "تريليونات", "كوادريليونات",
            "كوينتليونات", "سكستيليونات", "سبتيليونات", "أوكتيليونات",
            "نونيليونات", "ديسيليونات", "أندسيليونات", "دوديسيليونات",
            "تريديسيليونات", "كوادريسيليونات", "كوينتينيليونات"
        ]
        assert len(self.arabicAppendedGroup) == len(self.arabicGroup)
        assert len(self.arabicPluralGroups) == len(self.arabicGroup)
        assert len(self.arabicAppendedTwos) == len(self.arabicTwos)

    def number_to_arabic(self, arabic_prefix_text, arabic_suffix_text):
        self.arabicPrefixText = arabic_prefix_text
        self.arabicSuffixText = arabic_suffix_text
        self.extract_integer_and_decimal_parts()

    def extract_integer_and_decimal_parts(self):
        splits = re.split('\\.', str(self.number))

        self.integer_value = int(splits[0])
        if len(splits) > 1:
            self._decimalValue = int(self.decimal_value(splits[1]))
        else:
            self._decimalValue = 0

    def decimal_value(self, decimal_part):

        if self.partPrecision is not len(decimal_part):
            decimal_part_length = len(decimal_part)

            decimal_part_builder = decimal_part
            for i in range(0, self.partPrecision - decimal_part_length):
                decimal_part_builder += '0'
            decimal_part = decimal_part_builder

            if len(decimal_part) <= self.partPrecision:
                dec = len(decimal_part)
            else:
                dec = self.partPrecision
            result = decimal_part[0: dec]
        else:
            result = decimal_part

        # The following is useless (never happens)
        # for i in range(len(result), self.partPrecision):
        #     result += '0'
        return result

    def digit_feminine_status(self, digit, group_level):
        if group_level == -1:
            if self.isCurrencyPartNameFeminine:
                return self.arabicFeminineOnes[int(digit)]
            else:
                # Note: this never happens
                return self.arabicOnes[int(digit)]
        elif group_level == 0:
            if self.isCurrencyNameFeminine:
                return self.arabicFeminineOnes[int(digit)]
            else:
                return self.arabicOnes[int(digit)]
        else:
            return self.arabicOnes[int(digit)]

    def process_arabic_group(self, group_number, group_level,
                             remaining_number):
        tens = Decimal(group_number) % Decimal(100)
        hundreds = Decimal(group_number) / Decimal(100)
        ret_val = ""

        if int(hundreds) > 0:
            if tens == 0 and int(hundreds) == 2:
                ret_val = "{}".format(self.arabicAppendedTwos[0])
            else:
                ret_val = "{}".format(self.arabicHundreds[int(hundreds)])
                if ret_val != "" and tens != 0:
                    ret_val += " و "

        if tens > 0:
            if tens < 20:
                # if int(group_level) >= len(self.arabicTwos):
                #     raise OverflowError(self.errmsg_toobig %
                #                         (self.number, self.MAXVAL))
                assert int(group_level) < len(self.arabicTwos)
                if tens == 2 and int(hundreds) == 0 and group_level > 0:
                    pow = int(math.log10(self.integer_value))
                    if self.integer_value > 10 and pow % 3 == 0 and \
                            self.integer_value == 2 * (10 ** pow):
                        ret_val = "{}".format(
                            self.arabicAppendedTwos[int(group_level)])
                    else:
                        ret_val = "{}".format(
                            self.arabicTwos[int(group_level)])
                else:

                    if tens == 1 and group_level > 0 and hundreds == 0:
                        # Note: this never happens
                        # (hundreds == 0 only if group_number is 0)
                        ret_val += ""
                    elif (tens == 1 or tens == 2) and (
                            group_level == 0 or group_level == -1) and \
                            hundreds == 0 and remaining_number == 0:
                        # Note: this never happens (idem)
                        ret_val += ""
                    elif tens == 1 and group_level > 0:
                        ret_val += self.arabicGroup[int(group_level)]
                    else:
                        ret_val += self.digit_feminine_status(int(tens),
                                                              group_level)
            else:
                ones = tens % 10
                tens = (tens / 10) - 2
                if ones > 0:
                    ret_val += self.digit_feminine_status(ones, group_level)
                if ret_val != "" and ones != 0:
                    ret_val += " و "

                ret_val += self.arabicTens[int(tens)]

        return ret_val

    # We use this instead of built-in `abs` function,
    # because `abs` suffers from loss of precision for big numbers
    def abs(self, number):
        return number if number >= 0 else -number

    # We use this instead of `"{:09d}".format(number)`,
    # because the string conversion suffers from loss of
    # precision for big numbers
    def to_str(self, number):
        integer = int(number)
        if integer == number:
            return str(integer)
        decimal = round((number - integer) * 10**9)
        return str(integer) + "." + "{:09d}".format(decimal).rstrip("0")

    def convert(self, value):
        self.number = self.to_str(value)
        self.number_to_arabic(self.arabicPrefixText, self.arabicSuffixText)
        return self.convert_to_arabic()

    def convert_to_arabic(self):
        temp_number = Decimal(self.number)

        if temp_number == Decimal(0):
            return "صفر"

        decimal_string = self.process_arabic_group(self._decimalValue,
                                                   -1,
                                                   Decimal(0))
        ret_val = ""
        group = 0

        while temp_number > Decimal(0):

            temp_number_dec = Decimal(str(temp_number))
            try:
                number_to_process = int(temp_number_dec % Decimal(str(1000)))
            except decimal.InvalidOperation:
                decimal.getcontext().prec = len(
                    temp_number_dec.as_tuple().digits
                )
                number_to_process = int(temp_number_dec % Decimal(str(1000)))

            temp_number = int(temp_number_dec / Decimal(1000))

            group_description = \
                self.process_arabic_group(number_to_process,
                                          group,
                                          Decimal(floor(temp_number)))
            if group_description != '':
                if group > 0:
                    if ret_val != "":
                        ret_val = "{}و {}".format("", ret_val)
                    if number_to_process != 2 and number_to_process != 1:
                        # if group >= len(self.arabicGroup):
                        #     raise OverflowError(self.errmsg_toobig %
                        #                         (self.number, self.MAXVAL)
                        #     )
                        assert group < len(self.arabicGroup)
                        if number_to_process % 100 != 1:
                            if 3 <= number_to_process <= 10:
                                ret_val = "{} {}".format(
                                    self.arabicPluralGroups[group], ret_val)
                            else:
                                if ret_val != "":
                                    ret_val = "{} {}".format(
                                        self.arabicAppendedGroup[group],
                                        ret_val)
                                else:
                                    ret_val = "{} {}".format(
                                        self.arabicGroup[group], ret_val)

                        else:
                            ret_val = "{} {}".format(self.arabicGroup[group],
                                                     ret_val)
                ret_val = "{} {}".format(group_description, ret_val)
            group += 1
        formatted_number = ""
        if self.arabicPrefixText != "":
            formatted_number += "{} ".format(self.arabicPrefixText)
        formatted_number += ret_val
        if self.integer_value != 0:
            remaining100 = int(self.integer_value % 100)

            if remaining100 == 0:
                formatted_number += self.currency_unit[0]
            elif remaining100 == 1:
                formatted_number += self.currency_unit[0]
            elif remaining100 == 2:
                if self.integer_value == 2:
                    formatted_number += self.currency_unit[1]
                else:
                    formatted_number += self.currency_unit[0]
            elif 3 <= remaining100 <= 10:
                formatted_number += self.currency_unit[2]
            elif 11 <= remaining100 <= 99:
                formatted_number += self.currency_unit[3]
        if self._decimalValue != 0:
            formatted_number += " {} ".format(self.separator)
            formatted_number += decimal_string

        if self._decimalValue != 0:
            formatted_number += " "
            remaining100 = int(self._decimalValue % 100)

            if remaining100 == 0:
                formatted_number += self.currency_subunit[0]
            elif remaining100 == 1:
                formatted_number += self.currency_subunit[0]
            elif remaining100 == 2:
                formatted_number += self.currency_subunit[1]
            elif 3 <= remaining100 <= 10:
                formatted_number += self.currency_subunit[2]
            elif 11 <= remaining100 <= 99:
                formatted_number += self.currency_subunit[3]

        if self.arabicSuffixText != "":
            formatted_number += " {}".format(self.arabicSuffixText)

        return formatted_number

    def validate_number(self, number):
        if number >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig % (number, self.MAXVAL))
        return number

    def set_currency_prefer(self, currency):
        if currency == 'TND':
            self.currency_unit = CURRENCY_TND[0]
            self.currency_subunit = CURRENCY_TND[1]
            self.partPrecision = 3
        elif currency == 'EGP':
            self.currency_unit = CURRENCY_EGP[0]
            self.currency_subunit = CURRENCY_EGP[1]
            self.partPrecision = 2
        elif currency == 'KWD':
            self.currency_unit = CURRENCY_KWD[0]
            self.currency_subunit = CURRENCY_KWD[1]
            self.partPrecision = 2
        else:
            self.currency_unit = CURRENCY_SR[0]
            self.currency_subunit = CURRENCY_SR[1]
            self.partPrecision = 2

    def to_currency(self, value, currency='SR', prefix='', suffix=''):
        self.set_currency_prefer(currency)
        self.isCurrencyNameFeminine = False
        self.separator = "و"
        self.arabicOnes = ARABIC_ONES
        self.arabicPrefixText = prefix
        self.arabicSuffixText = suffix
        return self.convert(value=value)

    def to_ordinal(self, number, prefix=''):
        if number <= 19:
            return "{}".format(self.arabicOrdinal[number])
        if number < 100:
            self.isCurrencyNameFeminine = True
        else:
            self.isCurrencyNameFeminine = False
        self.currency_subunit = ('', '', '', '')
        self.currency_unit = ('', '', '', '')
        self.arabicPrefixText = prefix
        self.arabicSuffixText = ""
        return "{}".format(self.convert(self.abs(number)).strip())

    def to_year(self, value):
        value = self.validate_number(value)
        return self.to_cardinal(value)

    def to_ordinal_num(self, value):
        return self.to_ordinal(value).strip()

    def to_cardinal(self, number):
        self.isCurrencyNameFeminine = False
        number = self.validate_number(number)
        minus = ''
        if number < 0:
            minus = 'سالب '
        self.separator = ','
        self.currency_subunit = ('', '', '', '')
        self.currency_unit = ('', '', '', '')
        self.arabicPrefixText = ""
        self.arabicSuffixText = ""
        self.arabicOnes = ARABIC_ONES
        return minus + self.convert(value=self.abs(number)).strip()



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import print_function, unicode_literals


class Num2Word_ID():
    BASE = {0: [],
            1: ["satu"],
            2: ["dua"],
            3: ["tiga"],
            4: ["empat"],
            5: ["lima"],
            6: ["enam"],
            7: ["tujuh"],
            8: ["delapan"],
            9: ["sembilan"]}

    TENS_TO = {3: "ribu",
               6: "juta",
               9: "miliar",
               12: "triliun",
               15: "kuadriliun",
               18: "kuantiliun",
               21: "sekstiliun",
               24: "septiliun",
               27: "oktiliun",
               30: "noniliun",
               33: "desiliun"}

    errmsg_floatord = "Cannot treat float number as ordinal"
    errmsg_negord = "Cannot treat negative number as ordinal"
    errmsg_toobig = "Number is too large to convert to words (abs(%s) > %s)."
    MAXVAL = 10 ** 36

    def split_by_koma(self, number):
        return str(number).split('.')

    def split_by_3(self, number):
        """
        starting here, it groups the number by three from the tail
        '1234567' -> (('1',),('234',),('567',))
        :param number:str
        :rtype:tuple
        """
        blocks = ()
        length = len(number)

        if length < 3:
            blocks += ((number,),)
        else:
            len_of_first_block = length % 3

            if len_of_first_block > 0:
                first_block = number[0:len_of_first_block],
                blocks += first_block,

            for i in range(len_of_first_block, length, 3):
                next_block = (number[i:i + 3],),
                blocks += next_block

        return blocks

    def spell(self, blocks):
        """
        it adds the list of spelling to the blocks
        (
        ('1',),('034',)) -> (('1',['satu']),('234',['tiga', 'puluh', 'empat'])
        )
        :param blocks: tuple
        :rtype: tuple
        """
        word_blocks = ()
        first_block = blocks[0]
        if len(first_block[0]) == 1:
            if first_block[0] == '0':
                spelling = ['nol']
            else:
                spelling = self.BASE[int(first_block[0])]
        elif len(first_block[0]) == 2:
            spelling = self.puluh(first_block[0])
        else:
            spelling = (
                self.ratus(first_block[0][0]) + self.puluh(first_block[0][1:3])
                )

        word_blocks += (first_block[0], spelling),

        for block in blocks[1:]:
            spelling = self.ratus(block[0][0]) + self.puluh(block[0][1:3])
            block += spelling,
            word_blocks += block,

        return word_blocks

    def ratus(self, number):
        # it is used to spell
        if number == '1':
            return ['seratus']
        elif number == '0':
            return []
        else:
            return self.BASE[int(number)] + ['ratus']

    def puluh(self, number):
        # it is used to spell
        if number[0] == '1':
            if number[1] == '0':
                return ['sepuluh']
            elif number[1] == '1':
                return ['sebelas']
            else:
                return self.BASE[int(number[1])] + ['belas']
        elif number[0] == '0':
            return self.BASE[int(number[1])]
        else:
            return (
                self.BASE[int(number[0])] + ['puluh']
                + self.BASE[int(number[1])]
            )

    def spell_float(self, float_part):
        # spell the float number
        word_list = []
        for n in float_part:
            if n == '0':
                word_list += ['nol']
                continue
            word_list += self.BASE[int(n)]
        return ' '.join(['', 'koma'] + word_list)

    def join(self, word_blocks, float_part):
        """
        join the words by first join lists in the tuple
        :param word_blocks: tuple
        :rtype: str
        """
        word_list = []
        length = len(word_blocks) - 1
        first_block = word_blocks[0],
        start = 0

        if length == 1 and first_block[0][0] == '1':
            word_list += ['seribu']
            start = 1

        for i in range(start, length + 1, 1):
            word_list += word_blocks[i][1]
            if not word_blocks[i][1]:
                continue
            if i == length:
                break
            word_list += [self.TENS_TO[(length - i) * 3]]

        return ' '.join(word_list) + float_part

    def to_cardinal(self, number):
        if number >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig % (number, self.MAXVAL))
        minus = ''
        if number < 0:
            minus = 'min '
        float_word = ''
        n = self.split_by_koma(abs(number))
        if len(n) == 2:
            float_word = self.spell_float(n[1])
        return minus + self.join(self.spell(self.split_by_3(n[0])), float_word)

    def to_ordinal(self, number):
        self.verify_ordinal(number)
        out_word = self.to_cardinal(number)
        if out_word == "satu":
            return "pertama"
        return "ke" + out_word

    def to_ordinal_num(self, number):
        self.verify_ordinal(number)
        return "ke-" + str(number)

    def to_currency(self, value):
        return self.to_cardinal(value) + " rupiah"

    def to_year(self, value):
        return self.to_cardinal(value)

    def verify_ordinal(self, value):
        if not value == int(value):
            raise TypeError(self.errmsg_floatord % value)
        if not abs(value) == value:
            raise TypeError(self.errmsg_negord % value)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.
# Copyright (c) 2015, Blaz Bregar. All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .lang_EU import Num2Word_EU


class Num2Word_SL(Num2Word_EU):
    GIGA_SUFFIX = "ilijard"
    MEGA_SUFFIX = "ilijon"

    def setup(self):
        super(Num2Word_SL, self).setup()

        self.negword = "minus "
        self.pointword = "celih"
        self.errmsg_nonnum = "Only numbers may be converted to words."
        self.errmsg_toobig = (
            "Number is too large to convert to words (abs(%s) > %s)."
        )
        self.exclude_title = []

        self.mid_numwords = [(1000, "tisoč"), (900, "devetsto"),
                             (800, "osemsto"), (700, "sedemsto"),
                             (600, "šeststo"), (500, "petsto"),
                             (400, "štiristo"), (300, "tristo"),
                             (200, "dvesto"), (100, "sto"),
                             (90, "devetdeset"), (80, "osemdeset"),
                             (70, "sedemdeset"), (60, "šestdeset"),
                             (50, "petdeset"), (40, "štirideset"),
                             (30, "trideset")]
        self.low_numwords = ["dvajset", "devetnajst", "osemnajst",
                             "sedemnajst", "šestnajst", "petnajst",
                             "štirinajst", "trinajst", "dvanajst",
                             "enajst", "deset", "devet", "osem", "sedem",
                             "šest", "pet", "štiri", "tri", "dve", "ena",
                             "nič"]
        self.ords = {"ena": "prv",
                     "dve": "drug",
                     "tri": "tretj",
                     "štiri": "četrt",
                     "sedem": "sedm",
                     "osem": "osm",
                     "sto": "stot",
                     "tisoč": "tisoč",
                     "milijon": "milijont"
                     }
        self.ordflag = False

    def merge(self, curr, next):
        ctext, cnum, ntext, nnum = curr + next

        if ctext.endswith("dve") and self.ordflag and nnum <= 1000000:
            ctext = ctext[:len(ctext)-1] + "a"

        if ctext == "dve" and not self.ordflag and nnum < 1000000000:
            ctext = "dva"

        if (ctext.endswith("tri") or ctext.endswith("štiri")) and\
           nnum == 1000000 and not self.ordflag:
            if ctext.endswith("štiri"):
                ctext = ctext[:-1]
            ctext = ctext + "je"

        if cnum >= 20 and cnum < 100 and nnum == 2:
            ntext = "dva"

        if ctext.endswith("ena") and nnum >= 1000:
            ctext = ctext[0:-1]

        if cnum == 1:
            if nnum < 10**6 or self.ordflag:
                return next
            ctext = ""

        if nnum > cnum:
            if nnum >= 10**6:
                if self.ordflag:
                    ntext += "t"

                elif cnum == 2:
                    if ntext.endswith("d"):
                        ntext += "i"
                    else:
                        ntext += "a"

                elif 2 < cnum < 5:
                    if ntext.endswith("d"):
                        ntext += "e"
                    elif not ntext.endswith("d"):
                        ntext += "i"

                elif ctext.endswith("en"):
                    if ntext.endswith("d") or ntext.endswith("n"):
                        ntext += ""

                elif ctext.endswith("dve") and ntext.endswith("n"):
                    ctext = ctext[:-1] + "a"
                    ntext += "a"

                elif ctext.endswith("je") and ntext.endswith("n"):
                    ntext += "i"

                else:
                    if ntext.endswith("d"):
                        ntext += "a"
                    elif ntext.endswith("n"):
                        ntext += ""
                    elif ntext.endswith("d"):
                        ntext += "e"
                    else:
                        ntext += "ov"

            if nnum >= 10**2 and self.ordflag is False and ctext:
                ctext += " "

            val = cnum * nnum
        else:
            if nnum < 10 < cnum < 100:
                ntext, ctext = ctext, ntext + "in"
            elif cnum >= 10**2 and self.ordflag is False:
                ctext += " "
            val = cnum + nnum

        word = ctext + ntext
        return (word, val)

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        self.ordflag = True
        outword = self.to_cardinal(value)
        self.ordflag = False
        for key in self.ords:
            if outword.endswith(key):
                outword = outword[:len(outword) - len(key)] + self.ords[key]
                break
        return outword + "i"

    # Is this correct??
    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return str(value) + "."

    def to_currency(self, val, longval=True, old=False):
        if old:
            return self.to_splitnum(val, hightxt="evro/a/v",
                                    lowtxt="stotin/a/i/ov",
                                    jointxt="in", longval=longval)
        return super(Num2Word_SL, self).to_currency(val, jointxt="in",
                                                    longval=longval)

    def to_year(self, val, longval=True):
        if not (val//100) % 10:
            return self.to_cardinal(val)
        return self.to_splitnum(val, hightxt="hundert", longval=longval)



# -*- coding: utf-8 -*-
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .base import Num2Word_Base
from .utils import get_digits, splitbyx

ZERO = ('nula',)

ONES = {
    1: ('jeden', 'jeden', set()),
    2: ('dva', 'dve', {1, 3, 5, 7, 9}),
    3: ('tri', 'tri', set()),
    4: ('štyri', 'štyri', set()),
    5: ('päť', 'päť', set()),
    6: ('šesť', 'šesť', set()),
    7: ('sedem', 'sedem', set()),
    8: ('osem', 'osem', set()),
    9: ('deväť', 'deväť', set()),
}

TENS = {
    0: ('desať',),
    1: ('jedenásť',),
    2: ('dvanásť',),
    3: ('trinásť',),
    4: ('štrnásť',),
    5: ('pätnásť',),
    6: ('šestnásť',),
    7: ('sedemnásť',),
    8: ('osemnásť',),
    9: ('devätnásť',),
}

TWENTIES = {
    2: ('dvadsať',),
    3: ('tridsať',),
    4: ('štyridsať',),
    5: ('päťdesiat',),
    6: ('šesťdesiat',),
    7: ('sedemdesiat',),
    8: ('osemdesiat',),
    9: ('deväťdesiat',),
}

HUNDREDS = {
    1: ('sto',),
    2: ('dvesto',),
    3: ('tristo',),
    4: ('štyristo',),
    5: ('päťsto',),
    6: ('šesťsto',),
    7: ('sedemsto',),
    8: ('osemsto',),
    9: ('deväťsto',),
}

THOUSANDS = {
    1: ('tisíc', 'tisíc', 'tisíc'),  # 10^3
    2: ('milión', 'milióny', 'miliónov'),  # 10^6
    3: ('miliarda', 'miliardy', 'miliárd'),  # 10^9
    4: ('bilión', 'bilióny', 'biliónov'),  # 10^12
    5: ('biliarda', 'biliardy', 'biliárd'),  # 10^15
    6: ('trilión', 'trilióny', 'triliónov'),  # 10^18
    7: ('triliarda', 'triliardy', 'triliárd'),  # 10^21
    8: ('kvadrilión', 'kvadrilióny', 'kvadriliónov'),  # 10^24
    9: ('kvadriliarda', 'kvadriliardy', 'kvadriliárd'),  # 10^27
    10: ('kvintilión', 'kvintillióny', 'kvintiliónov'),  # 10^30
}


class Num2Word_SK(Num2Word_Base):
    CURRENCY_FORMS = {
        'EUR': (
            ('euro', 'eurá', 'eur'), ('cent', 'centy', 'centov')
        ),
    }

    def setup(self):
        self.negword = "mínus"
        self.pointword = "celých"

    def to_cardinal(self, number):
        n = str(number).replace(',', '.')
        if '.' in n:
            left, right = n.split('.')
            leading_zero_count = len(right) - len(right.lstrip('0'))
            decimal_part = ((ZERO[0] + ' ') * leading_zero_count +
                            self._int2word(int(right)))
            return u'%s %s %s' % (
                self._int2word(int(left)),
                self.pointword,
                decimal_part
            )
        else:
            return self._int2word(int(n))

    def pluralize(self, n, forms):
        if n == 1:
            form = 0
        elif 0 < n < 5:
            form = 1
        else:
            form = 2
        return forms[form]

    def to_ordinal(self, value):
        raise NotImplementedError()

    def _int2word(self, n):
        if n == 0:
            return ZERO[0]

        words = []
        chunks = list(splitbyx(str(n), 3))
        i = len(chunks)
        for x in chunks:
            i -= 1

            if x == 0:
                continue

            n1, n2, n3 = get_digits(x)

            word_chunk = []

            if n3 > 0:
                word_chunk.append(HUNDREDS[n3][0])

            if n2 > 1:
                word_chunk.append(TWENTIES[n2][0])

            if n2 == 1:
                word_chunk.append(TENS[n1][0])
            elif n1 > 0 and not (i > 0 and x == 1):
                if n2 == 0 and n3 == 0 and i in ONES[n1][2]:
                    word_chunk.append(ONES[n1][1])
                else:
                    word_chunk.append(ONES[n1][0])
            if i > 1 and word_chunk:
                word_chunk.append(' ')
            if i > 0:
                word_chunk.append(self.pluralize(x, THOUSANDS[i]))
            words.append(''.join(word_chunk))

        return ' '.join(words[:-1]) + ''.join(words[-1:])



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from . import lang_EU


class Num2Word_EN(lang_EU.Num2Word_EU):
    def set_high_numwords(self, high):
        max = 3 + 3 * len(high)
        for word, n in zip(high, range(max, 3, -3)):
            self.cards[10 ** n] = word + "illion"

    def setup(self):
        super(Num2Word_EN, self).setup()

        self.negword = "minus "
        self.pointword = "point"
        self.exclude_title = ["and", "point", "minus"]

        self.mid_numwords = [(1000, "thousand"), (100, "hundred"),
                             (90, "ninety"), (80, "eighty"), (70, "seventy"),
                             (60, "sixty"), (50, "fifty"), (40, "forty"),
                             (30, "thirty")]
        self.low_numwords = ["twenty", "nineteen", "eighteen", "seventeen",
                             "sixteen", "fifteen", "fourteen", "thirteen",
                             "twelve", "eleven", "ten", "nine", "eight",
                             "seven", "six", "five", "four", "three", "two",
                             "one", "zero"]
        self.ords = {"one": "first",
                     "two": "second",
                     "three": "third",
                     "four": "fourth",
                     "five": "fifth",
                     "six": "sixth",
                     "seven": "seventh",
                     "eight": "eighth",
                     "nine": "ninth",
                     "ten": "tenth",
                     "eleven": "eleventh",
                     "twelve": "twelfth"}

    def merge(self, lpair, rpair):
        ltext, lnum = lpair
        rtext, rnum = rpair
        if lnum == 1 and rnum < 100:
            return (rtext, rnum)
        elif 100 > lnum > rnum:
            return ("%s-%s" % (ltext, rtext), lnum + rnum)
        elif lnum >= 100 > rnum:
            return ("%s and %s" % (ltext, rtext), lnum + rnum)
        elif rnum > lnum:
            return ("%s %s" % (ltext, rtext), lnum * rnum)
        return ("%s, %s" % (ltext, rtext), lnum + rnum)

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        outwords = self.to_cardinal(value).split(" ")
        lastwords = outwords[-1].split("-")
        lastword = lastwords[-1].lower()
        try:
            lastword = self.ords[lastword]
        except KeyError:
            if lastword[-1] == "y":
                lastword = lastword[:-1] + "ie"
            lastword += "th"
        lastwords[-1] = self.title(lastword)
        outwords[-1] = "-".join(lastwords)
        return " ".join(outwords)

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return "%s%s" % (value, self.to_ordinal(value)[-2:])

    def to_year(self, val, suffix=None, longval=True):
        if val < 0:
            val = abs(val)
            suffix = 'BC' if not suffix else suffix
        high, low = (val // 100, val % 100)
        # If year is 00XX, X00X, or beyond 9999, go cardinal.
        if (high == 0
                or (high % 10 == 0 and low < 10)
                or high >= 100):
            valtext = self.to_cardinal(val)
        else:
            hightext = self.to_cardinal(high)
            if low == 0:
                lowtext = "hundred"
            elif low < 10:
                lowtext = "oh-%s" % self.to_cardinal(low)
            else:
                lowtext = self.to_cardinal(low)
            valtext = "%s %s" % (hightext, lowtext)
        return (valtext if not suffix
                else "%s %s" % (valtext, suffix))



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

to_19 = (u'không', u'một', u'hai', u'ba', u'bốn', u'năm', u'sáu',
         u'bảy', u'tám', u'chín', u'mười', u'mười một', u'mười hai',
         u'mười ba', u'mười bốn', u'mười lăm', u'mười sáu', u'mười bảy',
         u'mười tám', u'mười chín')
tens = (u'hai mươi', u'ba mươi', u'bốn mươi', u'năm mươi',
        u'sáu mươi', u'bảy mươi', u'tám mươi', u'chín mươi')
denom = ('',
         u'nghìn', u'triệu', u'tỷ', u'nghìn tỷ', u'trăm nghìn tỷ',
         'Quintillion', 'Sextillion', 'Septillion', 'Octillion', 'Nonillion',
         'Decillion', 'Undecillion', 'Duodecillion', 'Tredecillion',
         'Quattuordecillion', 'Sexdecillion', 'Septendecillion',
         'Octodecillion', 'Novemdecillion', 'Vigintillion')


class Num2Word_VI(object):

    def _convert_nn(self, val):
        if val < 20:
            return to_19[val]
        for (dcap, dval) in ((k, 20 + (10 * v)) for (v, k) in enumerate(tens)):
            if dval + 10 > val:
                if val % 10:
                    a = u'lăm'
                    if to_19[val % 10] == u'một':
                        a = u'mốt'
                    else:
                        a = to_19[val % 10]
                    if to_19[val % 10] == u'năm':
                        a = u'lăm'
                    return dcap + ' ' + a
                return dcap

    def _convert_nnn(self, val):
        word = ''
        (mod, rem) = (val % 100, val // 100)
        if rem > 0:
            word = to_19[rem] + u' trăm'
            if mod > 0:
                word = word + ' '
        if mod > 0 and mod < 10:
            if mod == 5:
                word = word != '' and word + u'lẻ năm' or word + u'năm'
            else:
                word = word != '' and word + u'lẻ ' \
                    + self._convert_nn(mod) or word + self._convert_nn(mod)
        if mod >= 10:
            word = word + self._convert_nn(mod)
        return word

    def vietnam_number(self, val):
        if val < 100:
            return self._convert_nn(val)
        if val < 1000:
            return self._convert_nnn(val)
        for (didx, dval) in ((v - 1, 1000 ** v) for v in range(len(denom))):
            if dval > val:
                mod = 1000 ** didx
                lval = val // mod
                r = val - (lval * mod)

                ret = self._convert_nnn(lval) + u' ' + denom[didx]
                if 99 >= r > 0:
                    ret = self._convert_nnn(lval) + u' ' + denom[didx] + u' lẻ'
                if r > 0:
                    ret = ret + ' ' + self.vietnam_number(r)
                return ret

    def number_to_text(self, number):
        number = '%.2f' % number
        the_list = str(number).split('.')
        start_word = self.vietnam_number(int(the_list[0]))
        final_result = start_word
        if len(the_list) > 1 and int(the_list[1]) > 0:
            end_word = self.vietnam_number(int(the_list[1]))
            final_result = final_result + ' phẩy ' + end_word
        return final_result

    def to_cardinal(self, number):
        return self.number_to_text(number)

    def to_ordinal(self, number):
        return self.to_cardinal(number)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from . import lang_EU


class Num2Word_DA(lang_EU.Num2Word_EU):
    GIGA_SUFFIX = "illiarder"
    MEGA_SUFFIX = "illioner"

    def setup(self):
        super(Num2Word_DA, self).setup()

        self.negword = "minus "
        self.pointword = "komma"
        self.exclude_title = ["og", "komma", "minus"]

        self.mid_numwords = [(1000, "tusind"), (100, "hundrede"),
                             (90, "halvfems"), (80, "firs"),
                             (70, "halvfjerds"), (60, "treds"),
                             (50, "halvtreds"), (40, "fyrre"), (30, "tredive")]
        self.low_numwords = ["tyve", "nitten", "atten", "sytten",
                             "seksten", "femten", "fjorten", "tretten",
                             "tolv", "elleve", "ti", "ni", "otte",
                             "syv", "seks", "fem", "fire", "tre", "to",
                             "et", "nul"]
        self.ords = {"nul": "nul",
                     "et": "f\xf8rste",
                     "to": "anden",
                     "tre": "tredje",
                     "fire": "fjerde",
                     "fem": "femte",
                     "seks": "sjette",
                     "syv": "syvende",
                     "otte": "ottende",
                     "ni": "niende",
                     "ti": "tiende",
                     "elleve": "ellevte",
                     "tolv": "tolvte",
                     "tretten": "trett",
                     "fjorten": "fjort",
                     "femten": "femt",
                     "seksten": "sekst",
                     "sytten": "sytt",
                     "atten": "att",
                     "nitten": "nitt",
                     "tyve": "tyv"}
        self.ordflag = False

    def merge(self, curr, next):
        ctext, cnum, ntext, nnum = curr + next
        if next[1] == 100 or next[1] == 1000:
            lst = list(next)
            lst[0] = 'et' + lst[0]
            next = tuple(lst)

        if cnum == 1:
            if nnum < 10 ** 6 or self.ordflag:
                return next
            ctext = "en"
        if nnum > cnum:
            if nnum >= 10 ** 6:
                ctext += " "
            val = cnum * nnum
        else:
            if cnum >= 100 and cnum < 1000:
                ctext += " og "
            elif cnum >= 1000 and cnum <= 100000:
                ctext += "e og "
            if nnum < 10 < cnum < 100:
                if nnum == 1:
                    ntext = "en"
                ntext, ctext = ctext, ntext + "og"
            elif cnum >= 10 ** 6:
                ctext += " "
            val = cnum + nnum
        word = ctext + ntext
        return (word, val)

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        self.ordflag = True
        outword = self.to_cardinal(value)
        self.ordflag = False
        for key in self.ords:
            if outword.endswith(key):
                outword = outword[:len(outword) - len(key)] + self.ords[key]
                break
        if value % 100 >= 30 and value % 100 <= 39 or value % 100 == 0:
            outword += "te"
        elif value % 100 > 12 or value % 100 == 0:
            outword += "ende"
        return outword

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        vaerdte = (0, 1, 5, 6, 11, 12)
        if value % 100 >= 30 and value % 100 <= 39 or value % 100 in vaerdte:
            return str(value) + "te"
        elif value % 100 == 2:
            return str(value) + "en"
        return str(value) + "ende"

    def to_currency(self, val, longval=True):
        if val // 100 == 1 or val == 1:
            ret = self.to_splitnum(val, hightxt="kr", lowtxt="\xf8re",
                                   jointxt="og", longval=longval)
            return "en " + ret[3:]
        return self.to_splitnum(val, hightxt="kr", lowtxt="\xf8re",
                                jointxt="og", longval=longval)

    def to_year(self, val, longval=True):
        if val == 1:
            return 'en'
        if not (val // 100) % 10:
            return self.to_cardinal(val)
        return self.to_splitnum(val, hightxt="hundrede", longval=longval)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.
# Copyright (c) 2017, Tufan Kaynak, Framras.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .base import Num2Word_Base


class Num2Word_TR(Num2Word_Base):
    def __init__(self):
        self.precision = 2
        self.negword = u"eksi"
        self.pointword = u"virgül"
        self.CURRENCY_UNIT = u"lira"
        self.CURRENCY_SUBUNIT = u"kuruş"
        self.errmsg_nonnum = u"Sadece sayılar yazıya çevrilebilir."
        self.errmsg_floatord = u"Tam sayı olmayan {} sıralamada kullanılamaz."
        self.errmsg_negord = u"Pozitif olmayan {} sıralamada kullanılamaz."
        self.errmsg_toobig = u"abs({}) sayı yazıya çevirmek için çok büyük. " \
                             u"Yazıya çevrilebilecek en büyük rakam {}."
        self.exclude_title = []
        self.DECIMAL_SIGN = ","
        self.ORDINAL_SIGN = "."
        self.ZERO = u"sıfır"
        self.CARDINAL_ONES = {
            "1": u"bir",
            "2": u"iki",
            "3": u"üç",
            "4": u"dört",
            "5": u"beş",
            "6": u"altı",
            "7": u"yedi",
            "8": u"sekiz",
            "9": u"dokuz"
        }
        self.ORDINAL_ONES = {
            "1": u"birinci",
            "2": u"ikinci",
            "3": u"üçüncü",
            "4": u"dördüncü",
            "5": u"beşinci",
            "6": u"altıncı",
            "7": u"yedinci",
            "8": u"sekizinci",
            "9": u"dokuzuncu"
        }
        self.CARDINAL_TENS = {
            "1": u"on",
            "2": u"yirmi",
            "3": u"otuz",
            "4": u"kırk",
            "5": u"elli",
            "6": u"altmış",
            "7": u"yetmiş",
            "8": u"seksen",
            "9": u"doksan"
        }
        self.ORDINAL_TENS = {
            "1": u"onuncu",
            "2": u"yirminci",
            "3": u"otuzuncu",
            "4": u"kırkıncı",
            "5": u"ellinci",
            "6": u"altmışıncı",
            "7": u"yetmişinci",
            "8": u"sekseninci",
            "9": u"doksanıncı"
        }
        self.HUNDREDS = {
            "2": u"iki",
            "3": u"üç",
            "4": u"dört",
            "5": u"beş",
            "6": u"altı",
            "7": u"yedi",
            "8": u"sekiz",
            "9": u"dokuz"
        }
        self.CARDINAL_HUNDRED = (u"yüz",)
        self.ORDINAL_HUNDRED = (u"yüzüncü",)
        self.CARDINAL_TRIPLETS = {
            1: u"bin",
            2: u"milyon",
            3: u"milyar",
            4: u"trilyon",
            5: u"katrilyon",
            6: u"kentilyon"
        }
        self.ORDINAL_TRIPLETS = {
            1: u"bininci",
            2: u"milyonuncu",
            3: u"milyarıncı",
            4: u"trilyonuncu",
            5: u"katrilyonuncu",
            6: u"kentilyon"
        }
        self.MAXVAL = (10 ** ((len(self.CARDINAL_TRIPLETS) + 1) * 3)) - 1

        self.integers_to_read = []
        self.total_triplets_to_read = 0
        self.total_digits_outside_triplets = 0
        self.order_of_last_zero_digit = 0

    def to_cardinal(self, value):
        wrd = ""
        is_cardinal = self.verify_cardinal(value)
        if not is_cardinal:
            return wrd

        if not int(value) == value:
            return self.to_cardinal_float(value)

        if str(value).startswith("-"):
            pre_word, value = self.negword, float(str(value)[1:])
        else:
            pre_word, value = "", float(value)

        self.to_splitnum(value)

        if self.order_of_last_zero_digit >= len(self.integers_to_read[0]):
            # number like 00 and all 0s and even more, raise error
            return "%s%s" % (pre_word, wrd)

        if self.total_triplets_to_read == 1:
            if self.total_digits_outside_triplets == 2:
                if self.order_of_last_zero_digit == 1:
                    # number like x0, read cardinal x0 and return
                    wrd += self.CARDINAL_TENS.get(
                        self.integers_to_read[0][0], ""
                    )
                    return "%s%s" % (pre_word, wrd)
                if self.order_of_last_zero_digit == 0:
                    # number like xy, read cardinal xy and return
                    wrd += self.CARDINAL_TENS.get(
                        self.integers_to_read[0][0], ""
                    )
                    wrd += self.CARDINAL_ONES.get(
                        self.integers_to_read[0][1], ""
                    )
                return "%s%s" % (pre_word, wrd)

            if self.total_digits_outside_triplets == 1:
                if self.order_of_last_zero_digit == 0:
                    # number like x, read cardinal x and return
                    wrd += self.CARDINAL_ONES.get(
                        self.integers_to_read[0][0], ""
                    )
                    if self.integers_to_read[0][0] == "0":
                        return self.ZERO
                    return "%s%s" % (pre_word, wrd)

            if self.total_digits_outside_triplets == 0:
                if self.order_of_last_zero_digit == 2:
                    # number like x00, read cardinal x00 and return
                    wrd += self.HUNDREDS.get(
                        self.integers_to_read[0][0], ""
                    )
                    wrd += self.CARDINAL_HUNDRED[0]
                    return "%s%s" % (pre_word, wrd)
                if self.order_of_last_zero_digit == 1:
                    # number like xy0, read cardinal xy0 and return
                    wrd += self.HUNDREDS.get(
                        self.integers_to_read[0][0], ""
                    )
                    wrd += self.CARDINAL_HUNDRED[0]
                    wrd += self.CARDINAL_TENS.get(
                        self.integers_to_read[0][1], ""
                    )
                    return "%s%s" % (pre_word, wrd)
                if self.order_of_last_zero_digit == 0:
                    # number like xyz, read cardinal xyz and return
                    wrd += self.HUNDREDS.get(
                        self.integers_to_read[0][0], ""
                    )
                    wrd += self.CARDINAL_HUNDRED[0]
                    wrd += self.CARDINAL_TENS.get(
                        self.integers_to_read[0][1], ""
                    )
                    wrd += self.CARDINAL_ONES.get(
                        self.integers_to_read[0][2], ""
                    )
                    return "%s%s" % (pre_word, wrd)

        if self.total_triplets_to_read >= 2:
            if self.total_digits_outside_triplets == 2:
                if self.order_of_last_zero_digit == len(
                        self.integers_to_read[0]) - 1:
                    # number like x0 and all 0s, read cardinal x0 0..0
                    #  and return
                    wrd += self.CARDINAL_TENS.get(
                        self.integers_to_read[0][0], ""
                    )
                    wrd += self.CARDINAL_TRIPLETS[
                        self.total_triplets_to_read - 1
                    ]
                    return "%s%s" % (pre_word, wrd)
                if self.order_of_last_zero_digit == len(
                        self.integers_to_read[0]) - 2:
                    # number like xy and all 0s, read cardinal xy 0..0
                    #  and return
                    wrd += self.CARDINAL_TENS.get(
                        self.integers_to_read[0][0], ""
                    )
                    wrd += self.CARDINAL_ONES.get(
                        self.integers_to_read[0][1], ""
                    )
                    wrd += self.CARDINAL_TRIPLETS[
                        self.total_triplets_to_read - 1
                    ]
                    return "%s%s" % (pre_word, wrd)
                if self.order_of_last_zero_digit < len(
                        self.integers_to_read[0]) - 2:
                    # number like xy and others, read cardinal xy n..n
                    #  and return
                    wrd += self.CARDINAL_TENS.get(
                        self.integers_to_read[0][0], ""
                    )
                    wrd += self.CARDINAL_ONES.get(
                        self.integers_to_read[0][1], ""
                    )
                    wrd += self.CARDINAL_TRIPLETS[
                        self.total_triplets_to_read - 1
                    ]

            if self.total_digits_outside_triplets == 1:
                if self.order_of_last_zero_digit == len(
                        self.integers_to_read[0]) - 1:
                    # number like x and all 0s, read cardinal x 0..0
                    #  and return
                    if not (self.total_triplets_to_read == 2 and
                            self.integers_to_read[0][0] == "1"):
                        wrd += self.CARDINAL_ONES.get(
                            self.integers_to_read[0][0], ""
                        )
                    wrd += self.CARDINAL_TRIPLETS[
                        self.total_triplets_to_read - 1
                    ]
                    return "%s%s" % (pre_word, wrd)
                if self.order_of_last_zero_digit < len(
                        self.integers_to_read[0]) - 1:
                    # number like x and others, read cardinal x n..n
                    #  and return
                    if not (self.total_triplets_to_read == 2 and
                            self.integers_to_read[0][0] == "1"):
                        wrd += self.CARDINAL_ONES.get(
                            self.integers_to_read[0][0], ""
                        )
                    wrd += self.CARDINAL_TRIPLETS[
                        self.total_triplets_to_read - 1
                    ]

            if self.total_digits_outside_triplets == 0:
                if self.order_of_last_zero_digit == len(
                        self.integers_to_read[0]) - 1:
                    # number like x00 and all 0s, read cardinal x00 0..0
                    #  and return
                    wrd += self.HUNDREDS.get(self.integers_to_read[0][0], "")
                    wrd += self.CARDINAL_HUNDRED[0]
                    wrd += self.CARDINAL_TRIPLETS[
                        self.total_triplets_to_read - 1
                    ]
                    return "%s%s" % (pre_word, wrd)
                if self.order_of_last_zero_digit == len(
                        self.integers_to_read[0]) - 2:
                    # number like xy0 and all 0s, read cardinal xy0 0..0
                    #  and return
                    wrd += self.HUNDREDS.get(
                        self.integers_to_read[0][0], ""
                    )
                    wrd += self.CARDINAL_HUNDRED[0]
                    wrd += self.CARDINAL_TENS.get(
                        self.integers_to_read[0][1], ""
                    )
                    wrd += self.CARDINAL_TRIPLETS[
                        self.total_triplets_to_read - 1
                    ]
                    return "%s%s" % (pre_word, wrd)
                if self.order_of_last_zero_digit == len(
                        self.integers_to_read[0]) - 3:
                    # number like xyz and all 0s, read cardinal xyz 0..0
                    #  and return
                    wrd += self.HUNDREDS.get(self.integers_to_read[0][0], "")
                    wrd += self.CARDINAL_HUNDRED[0]
                    wrd += self.CARDINAL_TENS.get(
                        self.integers_to_read[0][1], ""
                    )
                    wrd += self.CARDINAL_ONES.get(
                        self.integers_to_read[0][2], ""
                    )
                    wrd += self.CARDINAL_TRIPLETS[
                        self.total_triplets_to_read - 1
                    ]
                    return "%s%s" % (pre_word, wrd)
                if self.order_of_last_zero_digit < len(
                        self.integers_to_read[0]) - 3:
                    # number like xyz and all others, read cardinal xyz n..n
                    wrd += self.HUNDREDS.get(self.integers_to_read[0][0], "")
                    wrd += self.CARDINAL_HUNDRED[0]
                    wrd += self.CARDINAL_TENS.get(
                        self.integers_to_read[0][1], ""
                    )
                    if not (self.total_triplets_to_read == 2 and
                            self.integers_to_read[0][2] == "1"):
                        wrd += self.CARDINAL_ONES.get(
                            self.integers_to_read[0][2], ""
                        )
                    wrd += self.CARDINAL_TRIPLETS[
                        self.total_triplets_to_read - 1
                    ]

            for i in list(range(self.total_triplets_to_read - 1, 0, -1)):
                reading_triplet_order = self.total_triplets_to_read - i
                if self.total_digits_outside_triplets == 0:
                    last_read_digit_order = reading_triplet_order * 3
                else:
                    last_read_digit_order = (reading_triplet_order - 1) * 3 +\
                                            self.total_digits_outside_triplets

                if not self.integers_to_read[0][
                        last_read_digit_order: last_read_digit_order + 3
                ] == "000":
                    if not self.integers_to_read[0][
                        last_read_digit_order
                    ] == "0":
                        wrd += self.HUNDREDS.get(
                            self.integers_to_read[0][last_read_digit_order], ""
                        )
                        if self.order_of_last_zero_digit == len(
                                self.integers_to_read[0]) - (
                                last_read_digit_order) - 1:
                            if i == 1:
                                wrd += self.CARDINAL_HUNDRED[0]
                                return "%s%s" % (pre_word, wrd)
                            elif i > 1:
                                wrd += self.CARDINAL_HUNDRED[0]
                                wrd += self.CARDINAL_TRIPLETS[i - 1]
                                return "%s%s" % (pre_word, wrd)
                        else:
                            wrd += self.CARDINAL_HUNDRED[0]

                    if not self.integers_to_read[0][
                                last_read_digit_order + 1] == "0":
                        if self.order_of_last_zero_digit == len(
                                self.integers_to_read[0]) - (
                                last_read_digit_order) - 2:
                            if i == 1:
                                wrd += self.CARDINAL_TENS.get(
                                    self.integers_to_read[0][
                                        last_read_digit_order + 1], ""
                                )
                                return "%s%s" % (pre_word, wrd)
                            elif i > 1:
                                wrd += self.CARDINAL_TENS.get(
                                    self.integers_to_read[0][
                                        last_read_digit_order + 1], ""
                                )
                                wrd += self.CARDINAL_TRIPLETS[i - 1]
                                return "%s%s" % (pre_word, wrd)
                        else:
                            wrd += self.CARDINAL_TENS.get(
                                self.integers_to_read[0][
                                    last_read_digit_order + 1], ""
                            )

                    if not self.integers_to_read[0][
                                last_read_digit_order + 2] == "0":
                        if self.order_of_last_zero_digit == len(
                                self.integers_to_read[0]) - (
                                last_read_digit_order) - 3:
                            if i == 1:
                                wrd += self.CARDINAL_ONES.get(
                                    self.integers_to_read[0][
                                        last_read_digit_order + 2], ""
                                )
                                return "%s%s" % (pre_word, wrd)
                            if i == 2:
                                if not self.integers_to_read[0][
                                        last_read_digit_order:
                                        last_read_digit_order + 2
                                        ] == "00":
                                    wrd += self.CARDINAL_ONES.get(
                                        self.integers_to_read[0][
                                            last_read_digit_order + 2], ""
                                    )
                                elif not self.integers_to_read[0][
                                            last_read_digit_order + 2] == "1":
                                    wrd += self.CARDINAL_ONES.get(
                                        self.integers_to_read[0][
                                            last_read_digit_order + 2], ""
                                    )
                                wrd += self.CARDINAL_TRIPLETS[i - 1]
                                return "%s%s" % (pre_word, wrd)
                            if i > 2:
                                wrd += self.CARDINAL_ONES.get(
                                    self.integers_to_read[0][
                                        last_read_digit_order + 2], ""
                                )
                                wrd += self.CARDINAL_TRIPLETS[i - 1]
                                return "%s%s" % (pre_word, wrd)
                        else:
                            if not self.integers_to_read[0][
                                    last_read_digit_order:
                                    last_read_digit_order + 2
                            ] == "00":
                                wrd += self.CARDINAL_ONES.get(
                                    self.integers_to_read[0][
                                        last_read_digit_order + 2], ""
                                )
                            else:
                                if i == 2:
                                    if not self.integers_to_read[0][
                                           last_read_digit_order:
                                           last_read_digit_order + 2
                                    ] == "00":
                                        wrd += self.CARDINAL_ONES.get(
                                            self.integers_to_read[0][
                                                last_read_digit_order + 2], ""
                                        )
                                    elif not self.integers_to_read[0][
                                                last_read_digit_order + 2
                                    ] == "1":
                                        wrd += self.CARDINAL_ONES.get(
                                            self.integers_to_read[0][
                                                last_read_digit_order + 2], ""
                                        )

                    wrd += self.CARDINAL_TRIPLETS[i - 1]

        return "%s%s" % (pre_word, wrd)

    def to_cardinal_float(self, value):
        self.to_splitnum(value)
        wrd = ""
        wrd += self.pointword
        if len(self.integers_to_read[1]) >= 1:
            wrd += self.CARDINAL_TENS.get(self.integers_to_read[1][0], "")

        if len(self.integers_to_read[1]) == 2:
            wrd += self.CARDINAL_ONES.get(self.integers_to_read[1][1], "")

        if self.integers_to_read[0] == "0":
            wrd = self.ZERO + wrd
        else:
            wrd = self.to_cardinal(int(self.integers_to_read[0])) + wrd
        return wrd

    def verify_cardinal(self, value):
        iscardinal = True
        try:
            if not float(value) == value:
                iscardinal = False
        except (ValueError, TypeError):
            raise TypeError(self.errmsg_nonnum)
        if abs(value) >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig.format(value, self.MAXVAL))
        return iscardinal

    def verify_ordinal(self, value):
        isordinal = True
        try:
            if not int(value) == value:
                isordinal = False
            if not abs(value) == value:
                raise TypeError(self.errmsg_negord.format(value))
        except (ValueError, TypeError):
            raise TypeError(self.errmsg_nonnum)
        if abs(value) >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig.format(value, self.MAXVAL))
        return isordinal

    def to_ordinal(self, value):
        wrd = ""
        isordinal = self.verify_ordinal(value)
        if isordinal:
            self.to_splitnum(value)

            if self.order_of_last_zero_digit >= len(self.integers_to_read[0]):
                # number like 00 and all 0s and even more, raise error
                return wrd

            if self.total_triplets_to_read == 1:
                if self.total_digits_outside_triplets == 2:
                    if self.order_of_last_zero_digit == 1:
                        # number like x0, read ordinal x0 and return
                        wrd += self.ORDINAL_TENS.get(
                            self.integers_to_read[0][0], ""
                        )
                        return wrd
                    if self.order_of_last_zero_digit == 0:
                        # number like xy, read ordinal xy and return
                        wrd += self.CARDINAL_TENS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.ORDINAL_ONES.get(
                            self.integers_to_read[0][1], ""
                        )
                        return wrd

                if self.total_digits_outside_triplets == 1:
                    if self.order_of_last_zero_digit == 0:
                        # number like x, read ordinal x and return
                        wrd += self.ORDINAL_ONES.get(
                            self.integers_to_read[0][0], ""
                        )
                        if self.integers_to_read[0][0] == "0":
                            return u"sıfırıncı"
                        return wrd

                if self.total_digits_outside_triplets == 0:
                    if self.order_of_last_zero_digit == 2:
                        # number like x00, read ordinal x00 and return
                        wrd += self.HUNDREDS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.ORDINAL_HUNDRED[0]
                        return wrd
                    if self.order_of_last_zero_digit == 1:
                        # number like xy0, read ordinal xy0 and return
                        wrd += self.HUNDREDS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.CARDINAL_HUNDRED[0]
                        wrd += self.ORDINAL_TENS.get(
                            self.integers_to_read[0][1], ""
                        )
                        return wrd
                    if self.order_of_last_zero_digit == 0:
                        # number like xyz, read ordinal xyz and return
                        wrd += self.HUNDREDS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.CARDINAL_HUNDRED[0]
                        wrd += self.CARDINAL_TENS.get(
                            self.integers_to_read[0][1], ""
                        )
                        if not self.integers_to_read[0][2] == "0":
                            wrd += self.ORDINAL_ONES.get(
                                self.integers_to_read[0][2], ""
                            )
                        return wrd

            if self.total_triplets_to_read >= 2:
                if self.total_digits_outside_triplets == 2:
                    if self.order_of_last_zero_digit == len(
                            self.integers_to_read[0]) - 1:
                        # number like x0 and all 0s, read ordinal x0 0..0
                        #  and return
                        wrd += self.CARDINAL_TENS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.ORDINAL_TRIPLETS[
                            self.total_triplets_to_read - 1
                        ]
                        return wrd
                    if self.order_of_last_zero_digit == len(
                            self.integers_to_read[0]) - 2:
                        # number like xy and all 0s, read ordinal xy 0..0
                        #  and return
                        wrd += self.CARDINAL_TENS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.CARDINAL_ONES.get(
                            self.integers_to_read[0][1], ""
                        )
                        wrd += self.ORDINAL_TRIPLETS[
                            self.total_triplets_to_read - 1
                        ]
                        return wrd
                    if self.order_of_last_zero_digit < len(
                            self.integers_to_read[0]) - 2:
                        # number like xy and others, read cardinal xy n..n
                        #  and return
                        wrd += self.CARDINAL_TENS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.CARDINAL_ONES.get(
                            self.integers_to_read[0][1], ""
                        )
                        wrd += self.CARDINAL_TRIPLETS[
                            self.total_triplets_to_read - 1
                        ]

                if self.total_digits_outside_triplets == 1:
                    if self.order_of_last_zero_digit == len(
                            self.integers_to_read[0]) - 1:
                        # number like x and all 0s, read ordinal x 0..0
                        #  and return
                        if not (self.total_triplets_to_read == 2 and
                                self.integers_to_read[0][0] == "1"):
                            wrd += self.CARDINAL_ONES.get(
                                self.integers_to_read[0][0], ""
                            )
                        wrd += self.ORDINAL_TRIPLETS[
                            self.total_triplets_to_read - 1
                        ]
                        return wrd
                    if self.order_of_last_zero_digit < len(
                            self.integers_to_read[0]) - 1:
                        # number like x and others, read cardinal x n..n
                        #  and return
                        if not (self.total_triplets_to_read == 2 and
                                self.integers_to_read[0][0] == "1"):
                            wrd += self.CARDINAL_ONES.get(
                                self.integers_to_read[0][0], ""
                            )
                        wrd += self.CARDINAL_TRIPLETS[
                            self.total_triplets_to_read - 1
                        ]

                if self.total_digits_outside_triplets == 0:
                    if self.order_of_last_zero_digit == len(
                            self.integers_to_read[0]) - 1:
                        # number like x00 and all 0s, read ordinal x00 0..0
                        #  and return
                        wrd += self.HUNDREDS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.CARDINAL_HUNDRED[0]
                        wrd += self.ORDINAL_TRIPLETS[
                            self.total_triplets_to_read - 1
                        ]
                        return wrd
                    if self.order_of_last_zero_digit == len(
                            self.integers_to_read[0]) - 2:
                        # number like xy0 and all 0s, read ordinal xy0 0..0
                        #  and return
                        wrd += self.HUNDREDS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.CARDINAL_HUNDRED[0]
                        wrd += self.CARDINAL_TENS.get(
                            self.integers_to_read[0][1], ""
                        )
                        wrd += self.ORDINAL_TRIPLETS[
                            self.total_triplets_to_read - 1
                        ]
                        return wrd
                    if self.order_of_last_zero_digit == len(
                            self.integers_to_read[0]) - 3:
                        # number like xyz and all 0s, read ordinal xyz 0..0
                        #  and return
                        wrd += self.HUNDREDS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.CARDINAL_HUNDRED[0]
                        wrd += self.CARDINAL_TENS.get(
                            self.integers_to_read[0][1], ""
                        )
                        wrd += self.CARDINAL_ONES.get(
                            self.integers_to_read[0][2], ""
                        )
                        wrd += self.ORDINAL_TRIPLETS[
                            self.total_triplets_to_read - 1
                        ]
                        return wrd
                    if self.order_of_last_zero_digit < len(
                            self.integers_to_read[0]) - 3:
                        # number like xyz and all others, read cardinal
                        #  xyz n..n
                        wrd += self.HUNDREDS.get(
                            self.integers_to_read[0][0], ""
                        )
                        wrd += self.CARDINAL_HUNDRED[0]
                        wrd += self.CARDINAL_TENS.get(
                            self.integers_to_read[0][1], ""
                        )
                        if not (self.total_triplets_to_read == 2 and
                                self.integers_to_read[0][2] == "1"):
                            wrd += self.CARDINAL_ONES.get(
                                self.integers_to_read[0][2], ""
                            )
                        wrd += self.CARDINAL_TRIPLETS[
                            self.total_triplets_to_read - 1
                        ]

                for i in list(range(self.total_triplets_to_read - 1, 0, -1)):
                    reading_triplet_order = self.total_triplets_to_read - i
                    if self.total_digits_outside_triplets == 0:
                        last_read_digit_order = reading_triplet_order * 3
                    else:
                        last_read_digit_order = \
                            (reading_triplet_order - 1) * 3 + \
                            self.total_digits_outside_triplets

                    if not self.integers_to_read[0][
                           last_read_digit_order: last_read_digit_order + 3
                           ] == "000":
                        if not self.integers_to_read[0][
                            last_read_digit_order
                        ] == "0":
                            if not self.integers_to_read[0][
                                last_read_digit_order
                            ] == "1":
                                wrd += self.CARDINAL_ONES.get(
                                    self.integers_to_read[0][
                                        last_read_digit_order
                                    ], ""
                                )
                            if self.order_of_last_zero_digit == len(
                                    self.integers_to_read[0]) - (
                                    last_read_digit_order) - 1:
                                if i == 1:
                                    wrd += self.ORDINAL_HUNDRED[0]
                                    return wrd
                                elif i > 1:
                                    wrd += self.CARDINAL_HUNDRED[0]
                                    wrd += self.ORDINAL_TRIPLETS[i - 1]
                                    return wrd
                            else:
                                wrd += self.CARDINAL_HUNDRED[0]

                        if not self.integers_to_read[0][
                                    last_read_digit_order + 1
                        ] == "0":
                            if self.order_of_last_zero_digit == len(
                                    self.integers_to_read[0]) - (
                                    last_read_digit_order) - 2:
                                if i == 1:
                                    wrd += self.ORDINAL_TENS.get(
                                        self.integers_to_read[0][
                                            last_read_digit_order + 1], ""
                                    )
                                    return wrd
                                elif i > 1:
                                    wrd += self.CARDINAL_TENS.get(
                                        self.integers_to_read[0][
                                            last_read_digit_order + 1], ""
                                    )
                                    wrd += self.ORDINAL_TRIPLETS[i - 1]
                                    return wrd
                            else:
                                wrd += self.CARDINAL_TENS.get(
                                    self.integers_to_read[0][
                                        last_read_digit_order + 1], ""
                                )

                        if not self.integers_to_read[0][
                                    last_read_digit_order + 2
                        ] == "0":
                            if self.order_of_last_zero_digit == len(
                                    self.integers_to_read[0]) - (
                                    last_read_digit_order) - 3:
                                if i == 1:
                                    wrd += self.ORDINAL_ONES.get(
                                        self.integers_to_read[0][
                                            last_read_digit_order + 2], ""
                                    )
                                    return wrd
                                if i == 2:
                                    if not self.integers_to_read[0][
                                       last_read_digit_order:
                                            last_read_digit_order + 2] == "00":
                                        wrd += self.CARDINAL_ONES.get(
                                            self.integers_to_read[0][
                                                last_read_digit_order + 2], ""
                                        )
                                    elif not self.integers_to_read[0][
                                                last_read_digit_order + 2
                                    ] == "1":
                                        wrd += self.CARDINAL_ONES.get(
                                            self.integers_to_read[0][
                                                last_read_digit_order + 2], ""
                                        )
                                    wrd += self.ORDINAL_TRIPLETS[i - 1]
                                    return wrd
                                if i > 2:
                                    wrd += self.CARDINAL_ONES.get(
                                        self.integers_to_read[0][
                                            last_read_digit_order + 2], ""
                                    )
                                    wrd += self.ORDINAL_TRIPLETS[i - 1]
                                    return wrd
                            else:
                                if not self.integers_to_read[0][
                                   last_read_digit_order:
                                        last_read_digit_order + 2] == "00":
                                    wrd += self.CARDINAL_ONES.get(
                                        self.integers_to_read[0][
                                            last_read_digit_order + 2], ""
                                    )
                                else:
                                    if not self.integers_to_read[0][
                                       last_read_digit_order:
                                           last_read_digit_order + 2] == "00":
                                        wrd += self.CARDINAL_ONES.get(
                                            self.integers_to_read[0][
                                                last_read_digit_order + 2], ""
                                        )
                                    elif not self.integers_to_read[0][
                                            last_read_digit_order + 2] == "1":
                                        wrd += self.CARDINAL_ONES.get(
                                            self.integers_to_read[0][
                                                last_read_digit_order + 2], ""
                                        )

                        wrd += self.CARDINAL_TRIPLETS[i - 1]

        return wrd

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return "%s%s" % (value, self.to_ordinal(value)[-4:])

    def to_splitnum(self, val):
        float_digits = str(int(val * 10 ** self.precision))
        if not int(val) == 0:
            self.integers_to_read = [
                str(int(val)),
                float_digits[len(float_digits) - self.precision:]
            ]
        else:
            self.integers_to_read = [
                "0",
                "0" * (self.precision - len(float_digits)) +
                float_digits[len(float_digits) - self.precision:]
            ]
        if len(self.integers_to_read[0]) % 3 > 0:
            self.total_triplets_to_read = (len(self.integers_to_read[0]) // 3)\
                                          + 1
        elif len(self.integers_to_read[0]) % 3 == 0:
            self.total_triplets_to_read = len(self.integers_to_read[0]) // 3
        self.total_digits_outside_triplets = len(self.integers_to_read[0]) % 3

        okunacak = list(self.integers_to_read[0][::-1])
        self.order_of_last_zero_digit = 0
        found = 0
        for i in range(len(okunacak) - 1):
            if int(okunacak[i]) == 0 and found == 0:
                self.order_of_last_zero_digit = i + 1
            else:
                found = 1

    def to_currency(self, value):
        if int(value) == 0:
            return u"bedelsiz"
        valueparts = self.to_cardinal(value).split(self.pointword)
        if len(valueparts) == 1:
            return valueparts[0] + self.CURRENCY_UNIT
        if len(valueparts) == 2:
            return self.CURRENCY_UNIT.join(valueparts) + \
                   self.CURRENCY_SUBUNIT



from __future__ import division, print_function, unicode_literals

import math

from .lang_EU import Num2Word_EU

GENERIC_DOLLARS = ('dòlar', 'dòlars')
GENERIC_CENTS = ('centau', 'centaus')
CURRENCIES_UNA = (
    'SLL',
    'SEK',
    'NOK',
    'CZK',
    'DKK',
    'ISK',
    'SKK',
    'GBP',
    'CYP',
    'EGP',
    'FKP',
    'GIP',
    'LBP',
    'SDG',
    'SHP',
    'SSP',
    'SYP',
    'INR',
    'IDR',
    'LKR',
    'MUR',
    'NPR',
    'PKR',
    'SCR',
    'ESP',
    'TRY',
    'ITL',
)
CENTS_UNA = ('EGP', 'JOD', 'LBP', 'SDG', 'SSP', 'SYP')


class Num2Word_CA(Num2Word_EU):
    CURRENCY_FORMS = {
        'EUR': (('euro', 'euros'), ('cèntim', 'cèntims')),
        'ESP': (('pesseta', 'pessetes'), ('cèntim', 'cèntims')),
        'USD': (GENERIC_DOLLARS, ('centau', 'centaus')),
        'PEN': (('sol', 'sols'), ('cèntim', 'cèntims')),
        'CRC': (('colón', 'colons'), GENERIC_CENTS),
        'AUD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'CAD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'GBP': (('lliura', 'lliures'), ('penic', 'penics')),
        'RUB': (('ruble', 'rubles'), ('copec', 'copecs')),
        'SEK': (('corona', 'corones'), ('öre', 'öre')),
        'NOK': (('corona', 'corones'), ('øre', 'øre')),
        'PLN': (('zloty', 'zlotys'), ('grosz', 'groszy')),
        'MXN': (('peso', 'pesos'), GENERIC_CENTS),
        'RON': (('leu', 'lei'), ('ban', 'bani')),
        'INR': (('rupia', 'rupies'), ('paisa', 'paise')),
        'HUF': (('fòrint', 'fòrints'), ('fillér', 'fillérs')),
        'FRF': (('franc', 'francs'), ('cèntim', 'cèntims')),
        'CNY': (('iuan', 'iuans'), ('fen', 'jiao')),
        'CZK': (('corona', 'corones'), ('haléř', 'haléřů')),
        'NIO': (('córdoba', 'córdobas'), GENERIC_CENTS),
        'VES': (('bolívar', 'bolívars'), ('cèntim', 'cèntims')),
        'BRL': (('real', 'reals'), GENERIC_CENTS),
        'CHF': (('franc', 'francs'), ('cèntim', 'cèntims')),
        'JPY': (('ien', 'iens'), ('sen', 'sen')),
        'KRW': (('won', 'wons'), ('jeon', 'jeon')),
        'KPW': (('won', 'wons'), ('chŏn', 'chŏn')),
        'TRY': (('lira', 'lires'), ('kuruş', 'kuruş')),
        'ZAR': (('rand', 'rands'), ('cèntim', 'cèntims')),
        'KZT': (('tenge', 'tenge'), ('tin', 'tin')),
        'UAH': (('hrívnia', 'hrívnies'), ('kopiika', 'kopíok')),
        'THB': (('baht', 'bahts'), ('satang', 'satang')),
        'AED': (('dirham', 'dirhams'), ('fils', 'fulūs')),
        'AFN': (('afgani', 'afganis'), ('puli', 'puls')),
        'ALL': (('lek', 'lekë'), ('qqindarka', 'qindarkë')),
        'AMD': (('dram', 'drams'), ('luma', 'lumas')),
        'ANG': (('florí', 'florins'), ('cèntim', 'cèntims')),
        'AOA': (('kwanza', 'kwanzes'), ('cèntim', 'cèntims')),
        'ARS': (('peso', 'pesos'), GENERIC_CENTS),
        'AWG': (('florí', 'florins'), GENERIC_CENTS),
        'AZN': (('manat', 'manats'), ('qəpik', 'qəpik')),
        'BBD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'BDT': (('taka', 'taka'), ('poisha', 'poisha')),
        'BGN': (('lev', 'leva'), ('stotinka', 'stotinki')),
        'BHD': (('dinar', 'dinars'), ('fils', 'fulūs')),
        'BIF': (('franc', 'francs'), ('cèntim', 'cèntims')),
        'BMD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'BND': (GENERIC_DOLLARS, GENERIC_CENTS),
        'BOB': (('boliviano', 'bolivianos'), GENERIC_CENTS),
        'BSD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'BTN': (('ngultrum', 'ngultrums'), ('chetrum', 'chetrums')),
        'BWP': (('pula', 'pula'), ('thebe', 'thebe')),
        'BYN': (('ruble', 'rubles'), ('copec', 'copecs')),
        'BYR': (('ruble', 'rubles'), ('copec', 'copecs')),
        'BZD': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'CDF': (('franc', 'francs'), ('cèntim', 'cèntims')),
        'CLP': (('peso', 'pesos'), GENERIC_CENTS),
        'COP': (('peso', 'pesos'), GENERIC_CENTS),
        'CUP': (('peso', 'pesos'), GENERIC_CENTS),
        'CVE': (('escut', 'escuts'), GENERIC_CENTS),
        'CYP': (('lliura', 'lliures'), ('cèntim', 'cèntims')),
        'DJF': (('franc', 'francs'), ('cèntim', 'cèntims')),
        'DKK': (('corona', 'corones'), ('øre', 'øre')),
        'DOP': (('peso', 'pesos'), GENERIC_CENTS),
        'DZD': (('dinar', 'dinars'), ('cèntim', 'cèntims')),
        'ECS': (('sucre', 'sucres'), GENERIC_CENTS),
        'EGP': (('lliura', 'lliures'), ('piastre', 'piastres')),
        'ERN': (('nakfa', 'nakfes'), ('cèntim', 'cèntims')),
        'ETB': (('birr', 'birr'), ('cèntim', 'cèntims')),
        'FJD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'FKP': (('lliura', 'lliures'), ('penic', 'penics')),
        'GEL': (('lari', 'laris'), ('tetri', 'tetri')),
        'GHS': (('cedi', 'cedis'), ('pesewa', 'pesewas')),
        'GIP': (('lliura', 'lliures'), ('penic', 'penics')),
        'GMD': (('dalasi', 'dalasis'), ('butut', 'bututs')),
        'GNF': (('franc', 'francs'), ('cèntim', 'cèntims')),
        'GTQ': (('quetzal', 'quetzals'), GENERIC_CENTS),
        'GYD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'HKD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'HNL': (('lempira', 'lempires'), GENERIC_CENTS),
        'HRK': (('kuna', 'kuna'), ('lipa', 'lipa')),
        'HTG': (('gourde', 'gourdes'), ('cèntim', 'cèntims')),
        'IDR': (('rúpia', 'rúpies'), ('cèntim', 'cèntims')),
        'ILS': (('xéquel', 'xéquels'), ('agorà', 'agorot')),
        'IQD': (('dinar', 'dinars'), ('fils', 'fils')),
        'IRR': (('rial', 'rials'), ('dinar', 'dinars')),
        'ISK': (('corona', 'corones'), ('eyrir', 'aurar')),
        'ITL': (('lira', 'lires'), ('cèntim', 'cèntims')),
        'JMD': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'JOD': (('dinar', 'dinars'), ('piastra', 'piastres')),
        'KES': (('xiling', 'xílings'), ('cèntim', 'cèntims')),
        'KGS': (('som', 'som'), ('tyiyn', 'tyiyn')),
        'KHR': (('riel', 'riels'), ('cèntim', 'cèntims')),
        'KMF': (('franc', 'francs'), ('cèntim', 'cèntims')),
        'KWD': (('dinar', 'dinars'), ('fils', 'fils')),
        'KYD': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'LAK': (('kip', 'kips'), ('at', 'at')),
        'LBP': (('lliura', 'lliures'), ('piastra', 'piastres')),
        'LKR': (('rúpia', 'rúpies'), ('cèntim', 'cèntims')),
        'LRD': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'LSL': (('loti', 'maloti'), ('sente', 'lisente')),
        'LTL': (('lita', 'litai'), ('cèntim', 'cèntims')),
        'LYD': (('dinar', 'dinars'), ('dírham', 'dírhams')),
        'MAD': (('dírham', 'dirhams'), ('cèntim', 'cèntims')),
        'MDL': (('leu', 'lei'), ('ban', 'bani')),
        'MGA': (('ariary', 'ariary'), ('iraimbilanja', 'iraimbilanja')),
        'MKD': (('denar', 'denari'), ('deni', 'deni')),
        'MMK': (('kyat', 'kyats'), ('pya', 'pyas')),
        'MNT': (('tögrög', 'tögrög'), ('möngö', 'möngö')),
        'MOP': (('pataca', 'pataques'), ('avo', 'avos')),
        'MRO': (('ouguiya', 'ouguiya'), ('khoums', 'khoums')),
        'MRU': (('ouguiya', 'ouguiya'), ('khoums', 'khoums')),
        'MUR': (('rupia', 'rúpies'), ('cèntim', 'cèntims')),
        'MVR': (('rufiyaa', 'rufiyaa'), ('laari', 'laari')),
        'MWK': (('kwacha', 'kwacha'), ('tambala', 'tambala')),
        'MYR': (('ringgit', 'ringgits'), ('sen', 'sens')),
        'MZN': (('metical', 'meticals'), GENERIC_CENTS),
        'NAD': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'NGN': (('naira', 'naires'), ('kobo', 'kobos')),
        'NPR': (('rupia', 'rupies'), ('paisa', 'paises')),
        'NZD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'OMR': (('rial', 'rials'), ('baisa', 'baisa')),
        'PAB': (GENERIC_DOLLARS, ('centésimo', 'centésimos')),
        'PGK': (('kina', 'kina'), ('toea', 'toea')),
        'PHP': (('peso', 'pesos'), GENERIC_CENTS),
        'PKR': (('rupia', 'rupies'), ('paisa', 'paise')),
        'PLZ': (('zloty', 'zlotys'), ('grosz', 'groszy')),
        'PYG': (('guaraní', 'guaranís'), ('cèntim', 'cèntims')),
        'QAR': (('rial', 'rials'), ('dírham', 'dírhams')),
        'QTQ': (('quetzal', 'quetzals'), GENERIC_CENTS),
        'RSD': (('dinar', 'dinars'), ('para', 'para')),
        'RUR': (('ruble', 'rubles'), ('copec', 'copecs')),
        'RWF': (('franc', 'francs'), ('cèntim', 'cèntims')),
        'SAR': (('riyal', 'riyals'), ('hàl·lala', 'hàl·lalat')),
        'SBD': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'SCR': (('rupia', 'rupies'), ('cèntim', 'cèntims')),
        'SDG': (('lliura', 'lliures'), ('piastre', 'piastres')),
        'SGD': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'SHP': (('lliura', 'lliures'), ('penic', 'penics')),
        'SLL': (('leonE', 'leones'), ('cèntim', 'cèntims')),
        'SRD': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'SSP': (('lliura', 'lliures'), ('piastre', 'piastres')),
        'STD': (('dobra', 'dobrAs'), ('cèntim', 'cèntims')),
        'SVC': (('colón', 'colons'), GENERIC_CENTS),
        'SYP': (('lliura', 'lliures'), ('piastre', 'piastres')),
        'SZL': (('lilangeni', 'emalangeni'), ('cèntim', 'cèntims')),
        'TJS': (('somoni', 'somoni'), ('diram', 'diram')),
        'TMT': (('manat', 'manats'), ('teňňesi', 'teňňesi')),
        'TND': (('dinar', 'dinars'), ('mil·lim', 'mil·limat')),
        'TOP': (('paanga', 'paangas'), ('seniti', 'seniti')),
        'TTD': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'TWD': (('nou dòlar', 'nous dòlars'), ('fen', 'fen')),
        'TZS': (('xíling', 'xílings'), ('cèntim', 'cèntims')),
        'UGX': (('xíling', 'xílings'), ('cèntim', 'cèntims')),
        'UYU': (('peso', 'pesos'), ('centèsim', 'centèsims')),
        'UZS': (('som', 'som'), ('tiyin', 'tiyin')),
        'VND': (('dong', 'dongs'), ('xu', 'xu')),
        'VUV': (('vatu', 'vatus'), ('cèntim', 'cèntims')),
        'WST': (('tala', 'tala'), ('sene', 'sene')),
        'XAF': (('franc CFA', 'francs CFA'), ('cèntim', 'cèntims')),
        'XCD': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'XOF': (('franc CFA', 'francs CFA'), ('cèntim', 'cèntims')),
        'XPF': (('franc CFP', 'francs CFP'), ('cèntim', 'cèntims')),
        'YER': (('rial', 'rials'), ('fils', 'fils')),
        'YUM': (('dinar', 'dinars'), ('para', 'para')),
        'ZMW': (('kwacha', 'kwacha'), ('ngwee', 'ngwee')),
        'ZWL': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
        'ZWL': (GENERIC_DOLLARS, ('cèntim', 'cèntims')),
    }

    GIGA_SUFFIX = None
    MEGA_SUFFIX = "ilió"

    def setup(self):
        lows = ["quadr", "tr", "b", "m"]
        self.high_numwords = self.gen_high_numwords([], [], lows)
        self.negword = "menys "
        self.pointword = "punt"
        self.errmsg_nonnum = "type(%s) no és [long, int, float]"
        self.errmsg_floatord = "El float %s no pot ser tractat com un" \
            " ordinal."
        self.errmsg_negord = "El número negatiu %s no pot ser tractat" \
            " com un ordinal."
        self.errmsg_toobig = "abs(%s) ha de ser inferior a %s."
        self.gender_stem = "è"
        self.exclude_title = ["i", "menys", "punt"]

        self.mid_numwords = [
            (1000, "mil"),
            (100, "cent"),
            (90, "noranta"),
            (80, "vuitanta"),
            (70, "setanta"),
            (60, "seixanta"),
            (50, "cinquanta"),
            (40, "quaranta"),
            (30, "trenta"),
        ]
        self.low_numwords = [
            "vint-i-nou",
            "vint-i-vuit",
            "vint-i-set",
            "vint-i-sis",
            "vint-i-cinc",
            "vint-i-quatre",
            "vint-i-tres",
            "vint-i-dos",
            "vint-i-un",
            "vint",
            "dinou",
            "divuit",
            "disset",
            "setze",
            "quinze",
            "catorze",
            "tretze",
            "dotze",
            "onze",
            "deu",
            "nou",
            "vuit",
            "set",
            "sis",
            "cinc",
            "quatre",
            "tres",
            "dos",
            "un",
            "zero",
        ]
        self.mid_num = {
            1000: "mil",
            100: "cent",
            90: "noranta",
            80: "vuitanta",
            70: "setanta",
            60: "seixanta",
            50: "cinquanta",
            40: "quaranta",
            30: "trenta",
            20: "vint",
            10: "deu",
        }
        self.low_num = {
            0: "zero",
            1: "un",
            2: "dos",
            3: "tres",
            4: "quatre",
            5: "cinc",
            6: "sis",
            7: "set",
            8: "vuit",
            9: "nou",
            10: "deu",
            11: "onze",
            12: "dotze",
            13: "tretze",
            14: "catorze",
            15: "quinze",
            16: "setze",
            17: "disset",
            18: "divuit",
            19: "dinou",
            20: "vint",
            21: "vint-i-un",
            22: "vint-i-dos",
            23: "vint-i-tres",
            24: "vint-i-quatre",
            25: "vint-i-cinc",
            26: "vint-i-sis",
            27: "vint-i-set",
            28: "vint-i-vuit",
            29: "vint-i-nou",
        }
        self.ords = {
            1: "primer",
            2: "segon",
            3: "tercer",
            4: "quart",
            5: "cinqu",
            6: "sis",
            7: "set",
            8: "vuit",
            9: "nov",
            10: "des",
            11: "onz",
            12: "dotz",
            13: "tretz",
            14: "catorz",
            15: "quinz",
            16: "setz",
            17: "disset",
            18: "divuit",
            19: "dinov",
            20: "vint",
            30: "trent",
            40: "quarant",
            50: "cinquant",
            60: "seixant",
            70: "setant",
            80: "vuitant",
            90: "norant",
            100: "cent",
            200: "dos-cent",
            300: "tres-cent",
            400: "quatre-cent",
            500: "cinc-cent",
            600: "sis-cent",
            700: "set-cent",
            800: "vuit-cent",
            900: "nou-cent",
            1e3: "mil",
            1e6: "milion",
            1e9: "mil milion",
            1e12: "bilion",
            1e15: "mil bilion",
        }

        self.ords_2 = {1: "1r", 2: "2n", 3: "3r", 4: "4t"}
        self.ords_3 = {
            1: "unè",
            2: "dosè",
            3: "tresè",
            4: "quatrè",
            5: "cinquè",
            6: "sisè",
            7: "setè",
            8: "vuitè",
            9: "novè",
        }

    def merge(self, curr, next):
        ctext, cnum, ntext, nnum = curr + next
        if cnum == 1:
            if nnum < 1000000:
                return next
            ctext = "un"

        if nnum < cnum:
            if cnum < 100:
                return "%s-%s" % (ctext, ntext), cnum + nnum
            elif nnum == 1:
                return "%s %s" % (ctext, ntext), cnum + nnum
            elif cnum == 100:
                return "%s %s" % (ctext, ntext), cnum + nnum
            else:
                return "%s %s" % (ctext, ntext), cnum + nnum
        elif (not nnum % 1000000) and cnum > 1:
            ntext = ntext[:-3] + "lions"
        if nnum == 100:
            ntext += "s"
            ctext += "-"
        else:
            ntext = " " + ntext
        return (ctext + ntext, cnum * nnum)

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        if value == 0:
            text = ""
        elif value < 5:
            text = self.ords[value]
        elif value <= 20:
            text = "%s%s" % (self.ords[value], self.gender_stem)
        elif value <= 30:
            frac = value % 10
            text = "%s%s%s" % (self.ords[20], "-i-", self.ords_3[frac])
        elif value < 100:
            dec = (value // 10) * 10
            text = "%s%s%s%s" % (self.ords[dec], "a",
                                 "-", self.ords_3[value - dec])
        elif value == 1e2:
            text = "%s%s" % (self.ords[value], self.gender_stem)
        elif value < 2e2:
            cen = (value // 100) * 100
            text = "%s %s" % (self.ords[cen], self.to_ordinal(value - cen))
        elif value < 1e3:
            cen = (value // 100) * 100
            text = "%s%s %s" % (self.ords[cen], "s",
                                self.to_ordinal(value - cen))
        elif value == 1e3:
            text = "%s%s" % (self.ords[value], self.gender_stem)
        elif value < 1e6:
            dec = 1000 ** int(math.log(int(value), 1000))
            high_part, low_part = divmod(value, dec)
            cardinal = self.to_cardinal(high_part) if high_part != 1 else ""
            text = "%s %s %s" % (cardinal, self.ords[dec],
                                 self.to_ordinal(low_part))
        elif value < 1e18:
            dec = 1000 ** int(math.log(int(value), 1000))
            high_part, low_part = divmod(value, dec)
            cardinal = self.to_cardinal(high_part) if high_part != 1 else ""
            text = "%s%s%s %s" % (cardinal, self.ords[dec],
                                  self.gender_stem, self.to_ordinal(low_part))
        else:
            part1 = self.to_cardinal(value)
            text = "%s%s" % (part1[:-1], "onè")
        return text.strip()

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        if value not in self.ords_2:
            return "%s%s" % (value, "è" if self.gender_stem == "è" else "a")
        else:
            return self.ords_2[value]

    def to_currency(self, val, currency="EUR", cents=True,
                    separator=" amb", adjective=False):
        result = super(Num2Word_CA, self).to_currency(
            val, currency=currency, cents=cents,
            separator=separator, adjective=adjective
        )
        list_result = result.split(separator + " ")

        if currency in CURRENCIES_UNA:
            list_result[0] = list_result[0].replace("un", "una")
            list_result[0] = list_result[0].replace("dos", "dues")
            list_result[0] = list_result[0].replace("cents", "centes")

        list_result[0] = list_result[0].replace("vint-i-un", "vint-i-un")
        list_result[0] = list_result[0].replace(" i un", "-un")
        list_result[0] = list_result[0].replace("un", "un")

        if currency in CENTS_UNA:
            list_result[1] = list_result[1].replace("un", "una")
            list_result[1] = list_result[1].replace("dos", "dues")

        list_result[1] = list_result[1].replace("vint-i-un", "vint-i-una")

        list_result[1] = list_result[1].replace("un", "un")

        result = (separator + " ").join(list_result)

        return result
